import sys
from threading import Thread, Event, Condition
from struct import pack, unpack
from codecs import encode

import logging
import re
import pygdbmi.gdbcontroller

import parse
if sys.version_info < (3, 0):
    import Queue as queue
    # __class__ = instance.__class__
else:
    import queue

from avatar2.archs.arm import ARM
from avatar2.targets import TargetStates
from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage, SyscallCatchedMessage

GDB_PROT_DONE = 'done'
GDB_PROT_CONN = 'connected'
GDB_PROT_RUN = 'running'


class GDBResponseListener(Thread):
    """
    This class creates objects waiting for responses from the gdb-process
    Depending whether a synchronous or asynchronous message is received,
    it is either put in a synchronous dictionary or parsed/lifted
    to an AvatarMessage and added to the Queue of the according target
    """

    def __init__(self, gdb_protocol, gdb_controller, avatar_queue,
                 avatar_fast_queue,  origin=None):
        super(GDBResponseListener, self).__init__()
        self._protocol = gdb_protocol
        self._token = -1
        self._async_responses = queue.Queue() if avatar_queue is None \
            else avatar_queue
        self._async_fast_responses = queue.Queue() if avatar_fast_queue is None\
            else avatar_fast_queue
        self._sync_responses = {}
        self._gdb_controller = gdb_controller
        self._gdb = gdb_protocol
        self._close = Event()
        self._closed = Event()
        self._close.clear()
        self._closed.clear()
        self._sync_responses_cv = Condition()
        self._last_exec_token = 0
        self._origin = origin
        self._console_output = None
        self._console_enable = False
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def get_token(self):
        """Gets a token for a synchronous request
        :returns: An (integer) token
        """
        self._token += 1
        return self._token

    def parse_async_notify(self, response):
        """
        This functions converts gdb notify messages to an avatar message

        :param response: A pygdbmi response dictonary
        :returns:        An avatar message
        """

        # Make sure this is a notify-response
        if response['type'] != 'notify':
            raise RuntimeError()

        msg = response['message']
        payload = response['payload']
        avatar_msg = None

        self.log.debug("Received Message: %s", msg)

        if msg.startswith('thread'):
            if msg == 'thread-group-exited':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.EXITED)
        elif msg.startswith('tsv'):
            pass  # likewise tracing
        elif msg.startswith('library'):
            pass  # library loading not supported yet
        elif msg == 'breakpoint-modified':
            pass  # ignore breakpoint modified for now
        elif msg == 'breakpoint-created':
            pass  # ignore breakpoint created for now
        elif msg == 'memory-changed':
            pass  # ignore changed memory for now
        elif msg == 'stopped':
            if payload.get('reason') == 'breakpoint-hit':
                avatar_msg = BreakpointHitMessage(self._origin, int(payload['bkptno']),
                                                  int(payload['frame']['addr'], 16))
            elif payload.get('reason') == 'exited-normally':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.EXITED)
            elif payload.get('reason') == 'end-stepping-range':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.STOPPED)
            elif payload.get('reason') == 'signal-received':
                if payload.get('signal-name') == 'SIGSEGV':
                    avatar_msg = UpdateStateMessage(
                        self._origin, TargetStates.EXITED)
                elif payload.get('signal-name') == 'SIGTRAP':
                    avatar_msg = BreakpointHitMessage(self._origin, -1,
                                                      int(payload['frame']['addr'], 16))
                else:
                    avatar_msg = UpdateStateMessage(
                        self._origin, TargetStates.STOPPED)
            elif payload.get('reason') == 'watchpoint-trigger':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.STOPPED)
            elif payload.get('reason') == 'access-watchpoint-trigger':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.STOPPED)
            elif payload.get('reason') == 'read-watchpoint-trigger':
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.STOPPED)
            elif payload.get('reason') == 'syscall-entry':
                avatar_msg = SyscallCatchedMessage(self._origin, int(payload['bkptno']),
                                                  int(payload['frame']['addr'], 16), 'entry')
            elif payload.get('reason') == 'syscall-return':
                avatar_msg = SyscallCatchedMessage(self._origin, int(payload['bkptno']),
                                                  int(payload['frame']['addr'], 16), 'return')
            elif payload.get('reason') is not None:
                self.log.critical("Target stopped with unknown reason: %s" %
                                  payload['reason'])
                # raise RuntimeError
            else:
                avatar_msg = UpdateStateMessage(
                    self._origin, TargetStates.STOPPED)
        elif msg == 'running':
            avatar_msg = UpdateStateMessage(self._origin, TargetStates.RUNNING)

        else:
            self.log.critical('Catched unknown async message: %s' % response)

        return avatar_msg

    def parse_async_response(self, response):
        """
        This functions converts a async gdb/mi message to an avatar message

        :param response: A pygdbmi response dictonary
        """

        if response['type'] == 'console':
            self.collect_console_output(response)
        elif response['type'] == 'log':
            pass  # TODO: implement handler for log messages
        elif response['type'] == 'target':
            pass  # TODO: implement handler for target messages
        elif response['type'] == 'output':
            pass  # TODO: implement handler for output messages
        elif response['type'] == 'notify':
            return self.parse_async_notify(response)


        else:
            raise Exception("GDBProtocol got unexpected response of type %s" %
                            response['type'])

    def get_async_response(self, timeout=0):
        return self._async_responses.get(timeout=timeout)

    def get_sync_response(self, token, timeout=5):
        for x in range(timeout * 2):
            self._sync_responses_cv.acquire()
            ret = self._sync_responses.pop(token, None)
            if ret is None:
                self._sync_responses_cv.wait(timeout=0.5)

            self._sync_responses_cv.release()
            if ret is not None:
                return ret
        raise TimeoutError()

    def run(self):
        while 1:
            if self._close.is_set():
                break

            try:
                responses = self._gdb_controller.get_gdb_response(
                    timeout_sec=0.01
                )
            except:
                continue

            for response in responses:
                if response.get('token', None) is not None:
                    self._sync_responses_cv.acquire()
                    self._sync_responses[response['token']] = response
                    self._sync_responses_cv.notifyAll()
                    self._sync_responses_cv.release()
                else:
                    avatar_msg = self.parse_async_response(response)
                    if avatar_msg is not None:
                        self.log.debug("Parsed an avatar_msg %s", avatar_msg)
                        if self._gdb._async_message_handler is not None:
                            self._gdb._async_message_handler(avatar_msg)
                        else:
                            if isinstance(avatar_msg, UpdateStateMessage):
                                self._async_fast_responses.put(avatar_msg)
                            else:
                                self._async_responses.put(avatar_msg)
        self._closed.set()

    def stop(self):
        """Stops the listening thread. Useful for teardown of the target"""
        self._close.set()
        self._closed.wait()

    def start_console_collection(self):
        self._console_output = ""
        self._console_enable = True

    def stop_console_collection(self):
        self._console_enable = False

    def collect_console_output(self, msg):
        if self._console_enable:
            self._console_output += '\n'
            self._console_output += msg['payload']

class GDBProtocol(object):
    """Main class for the gdb communication protocol
    :ivar gdb_executable: the path to the gdb which should be executed
    :ivar arch:     the architecture
    :ivar additional_args: additional arguments for gdb
    :ivar avatar:   the avatar object
    :ivar origin:   the target utilizing this protocol
    """

    def __init__(
            self,
            gdb_executable="gdb",
            arch=ARM,
            additional_args=[],
            async_message_handler=None,
            avatar=None,
            origin=None,
            enable_init_files=False,
            binary=None,
            local_arguments=None,
            verbose=False):
        self._async_message_handler = async_message_handler
        self._arch = arch
        self._register_mapping = dict(arch.registers)

        gdb_args = []
        if not enable_init_files:
            gdb_args += ['--nx']
        gdb_args += ['--quiet', '--interpreter=mi2']
        gdb_args += additional_args
        if binary is not None:
            gdb_args += ['--args', binary]
            if local_arguments is not None:
                gdb_args += [local_arguments]


        if sys.version_info <= (3, 6):
            self._gdbmi = pygdbmi.gdbcontroller.GdbController(
                gdb_path=gdb_executable,
                gdb_args=gdb_args,
                verbose=verbose)  # set to True for debugging
        else:
            self._gdbmi = pygdbmi.gdbcontroller.GdbController(
                command=[gdb_executable] + gdb_args,
                time_to_check_for_additional_output_sec=0)
        queue = avatar.queue if avatar is not None else None
        fast_queue = avatar.fast_queue if avatar is not None else None
        self._communicator = GDBResponseListener(
            self, self._gdbmi, queue, fast_queue, origin)
        self._communicator.daemon = True
        self._communicator.start()
        self._origin = origin
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def __del__(self):
        self.shutdown()

    def shutdown(self):
        if self._communicator is not None:
            self._communicator.stop()
            self._communicator = None
        if self._gdbmi is not None:
            self._gdbmi.exit()
            self._gdbmi = None

    def _sync_request(self, request, rexpect, timeout=5):
        """ Generic method to send a syncronized request

        :param request: the request as list
        :param rexpect: the expected response type
        :returns: Whether a response of type rexpect was received
        :returns: The response
        """

        token = self._communicator.get_token()
        request = [request] if isinstance(request, str) else request

        req = str(token) + ' '.join(request)
        self.log.debug("Sending request: %s" % req)

        self._gdbmi.write(req, read_response=False, timeout_sec=0)
        try:
            response = self._communicator.get_sync_response(token, timeout=timeout)
            ret = True if response['message'] == rexpect else False
        except:
            response = None
            ret = None
        return ret, response

    def set_abi(self, abi):
        req = ['-gdb-set', self._arch.gdb_name, 'abi', abi]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)

        if not ret:
            self.log.critical(
                "Unable to set abi to %s, received response: %s" %
                (abi, resp))
            raise Exception("GDBProtocol was unable to set ABI")

        return ret

    def remote_connect(self, ip='127.0.0.1', port=3333):
        """
        connect to a remote gdb server

        :param ip: ip of the remote gdb-server (default: localhost)
        :param port: port of the remote gdb-server (default: port)
        :returns: True on successful connection
        """

        req = ['-gdb-set', 'target-async', 'on']
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if not ret:
            self.log.critical(
                "Unable to set GDB/MI to async, received response: %s" %
                resp)
            raise Exception("GDBProtocol was unable to switch to async")

        req = ['-gdb-set', 'architecture', self._arch.gdb_name]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)


        if not ret:
            self.log.critical(
                "Unable to set architecture, received response: %s" %
                resp)
            raise Exception(("GDBProtocol was unable to set the architecture\n"
                             "Did you select the right gdb_executable?"))

        # if we are on ARM, set abi to AAPPCS to avoid bugs due to
        # fp-derefencation (https://github.com/avatartwo/avatar2/issues/19)
        if self._arch.gdb_name == 'arm':
            self.set_abi('AAPCS')

        if hasattr(self._arch, 'endian'):
            req = ['-gdb-set', 'endian', self._arch.endian]
            ret, resp = self._sync_request(req, GDB_PROT_DONE)
            if not ret:
                self.log.critical(
                    "Unable to set endianness, received response: %s" %
                    resp)
                raise Exception("GDBProtocol was unable to set endianness")

        req = ['-target-select', 'remote', '%s:%d' % (ip, int(port))]
        ret, resp = self._sync_request(req, GDB_PROT_CONN)

        self.log.debug(
            "Attempted to connect to target. Received response: %s" %
            resp)
        if not ret:
            self.log.critical("GDBProtocol was unable to connect to remote target")
            raise Exception("GDBProtocol was unable to connect")

        self.update_target_regs()

        return ret

    def remote_connect_serial(self, device='/dev/ttyACM0', baud_rate=38400,
                              parity='none'):
        """
        connect to a remote gdb server through a serial device

        :param device: file representing the device (default: /dev/ttyACM0)
        :param baud_rate: baud_rate of the serial device (default: 38400)
        :param parity: parity of the serial link (default no parity)
        :returns: True on successful connection
        """

        req = ['-gdb-set', 'architecture', self._arch.gdb_name]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if not ret:
            self.log.critical(
                "Unable to set architecture, received response: %s" %
                resp)
            raise Exception(("GDBProtocol was unable to set the architecture\n"
                             "Did you select the right gdb_executable?"))

        if parity not in ['none', 'even', 'odd']:
            self.log.critical("Parity must be none, even or odd")
            raise Exception("Cannot set parity to %s" % parity)

        req = ['-gdb-set', 'mi-async', 'on']
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if not ret:
            self.log.critical(
                "Unable to set GDB/MI to async, received response: %s" %
                resp)
            raise Exception("GDBProtocol was unable to connect")

        req = ['-gdb-set', 'serial', 'parity', '%s' % parity]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if not ret:
            self.log.critical("Unable to set parity")
            raise Exception("GDBProtocol was unable to set parity")

        req = ['-gdb-set', 'serial', 'baud', '%i' % baud_rate]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if not ret:
            self.log.critical("Unable to set baud rate")
            raise Exception("GDBProtocol was unable to set Baudrate")

        req = ['-target-select', 'remote', '%s' % device]
        ret, resp = self._sync_request(req, GDB_PROT_CONN)

        self.log.debug(
            "Attempted to connect to target. Received response: %s" %
            resp)

        self.update_target_regs()

        return ret

    def remote_disconnect(self):
        """
        disconnects from remote target
        """

        ret, resp = self._sync_request('-target-disconnect', GDB_PROT_DONE)

        self.log.debug(
            "Attempted to disconnect from target. Received response: %s" %
            resp)
        return ret

    def get_register_names(self):
        """fetch all register names
        :returns:  a list with all registers names, in order as known to gdb
        """

        ret, resp = self._sync_request(
            '-data-list-register-names', GDB_PROT_DONE)

        self.log.debug(
            "Attempted to obtain register names. Received response: %s" % resp)
        return resp['payload']['register-names'] if ret else None

    def update_target_regs(self):
        """
        This function will try to update the TargetRegs based on the list of
        registers known to gdb.
        """
        if hasattr(self._origin, 'regs'):
            regs = self.get_register_names()
            regs_dict = dict([(r,i) for i, r in enumerate(regs) if r != ''])
            self._origin.regs._update(regs_dict)


    def set_breakpoint(self, line,
                       hardware=False,
                       temporary=False,
                       regex=False,
                       condition=None,
                       ignore_count=0,
                       thread=0,
                       pending=False):
        """Inserts a breakpoint

        :param bool hardware: Hardware breakpoint
        :param bool temporary:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with specified condition
                              Note: This allows only a single condition. For more
                                    complex ones, please use set_break_condition()
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        :returns:             The number of the breakpoint
        """
        cmd = ["-break-insert"]
        if temporary:
            cmd.append("-t")
        if hardware:
            cmd.append("-h")
        if regex:
            assert ((not temporary) and (not condition) and (not ignore_count))
            cmd.append("-r")
            cmd.append(str(regex))
        if condition:
            cmd.append("-c")
            cmd.append(str(condition))
        if ignore_count:
            cmd.append("-i")
            cmd.append("%d" % ignore_count)
        if thread:
            cmd.append("-p")
            cmd.append("%d" % thread)
        if pending:
            cmd.append("-f")

        if isinstance(line, int):
            cmd.append("*0x%x" % line)
        else:
            cmd.append(str(line))

        ret, resp = self._sync_request(cmd, GDB_PROT_DONE)
        self.log.debug("Attempted to set breakpoint. Received response: %s" % resp)
        return int(resp['payload']['bkpt']['number']) if ret else -1

    def set_break_condition(self, bp_no, condition='1'):
        """
        Modifies the break condition for a given break-point
        :param bp_no: the breakpoint whose condition should be modified
        :param condition: The condition string to use.
                          The default value, '1', will make the bp unconditional
        """
        cmd = ["-break-condition", str(bp_no), condition]
        ret, resp = self._sync_request(cmd, GDB_PROT_DONE)
        self.log.debug("Attempted to set break-condition. Received response: %s" % resp)
        return ret

    def set_watchpoint(self, variable, write=True, read=False):
        cmd = ["-break-watch"]
        if read is True and write is True:
            cmd.append("-a")
        elif read is True:
            cmd.append("-r")
        elif write is True:
            pass
        else:
            raise ValueError("At least one read and write must be True")

        if isinstance(variable, int):
            cmd.append("*0x%x" % variable)
        else:
            cmd.append(str(variable))

        ret, resp = self._sync_request(cmd, GDB_PROT_DONE)
        self.log.debug("Attempted to set watchpoint. Received response: %s" % resp)

        if ret:
            # The payload contains different keys according to the
            # type of the watchpoint which has been set.
            # The possible keys are: [(hw-)][ar]wpt
            for k in resp['payload'].keys():
                if k.endswith('wpt'):
                    break
            else:
                return -1
            return int(resp['payload'][k]['number'])
        else:
            return -1

    def set_syscall_cachpoint(self, syscall):
        '''
        Set's up a syscall catchpoint.
        :param syscall: the syscall to catch; can be either its name or no
        '''
        # this command is not exported via gdb-mi, hence we need manual parsing
        ret = self.console_command('catch syscall %s' % syscall)
        if ret[0] is not True:
            return ret[0]
        # assume return message is in format of
        # '"\nCatchpoint 1 (syscall 'write' [4])\\n") '
        expected_bp_num = ret[1].split()[1]
        if not expected_bp_num.isdigit():
            self.log.warning("Couldn't extract bp_num for catchpoint!")
            return True
        return int(expected_bp_num)




    def remove_breakpoint(self, bkpt):
        """Deletes a breakpoint"""
        ret, resp = self._sync_request(
            ['-break-delete', str(bkpt)], GDB_PROT_DONE)

        self.log.debug(
            "Attempted to delete breakpoint. Received response: %s" %
            resp)
        return ret

    def write_memory(self, address, wordsize, val, num_words=1, raw=False):
        """Writes memory

        :param address:   Address to write to
        :param wordsize:  the size of the write (1, 2, 4 or 8)
        :param val:       the written value
        :type val:        int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """
        num2fmt = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}

        max_write_size = 0x100

        if raw:
            if not len(val):
                raise ValueError("val had zero length")
            for i in range(0, len(val), max_write_size):
                write_val = encode(val[i:max_write_size + i], 'hex_codec').decode('ascii')
                ret, resp = self._sync_request(
                    ["-data-write-memory-bytes", str(address + i), write_val],
                    GDB_PROT_DONE)

        else:
            fmt = '<%d%s' % (num_words, num2fmt[wordsize])
            if num_words == 1:
                contents = pack(fmt, val)
            else:
                contents = pack(fmt, *val)

            hex_contents = encode(contents, 'hex_codec').decode('ascii')
            ret, resp = self._sync_request(
                ["-data-write-memory-bytes", str(address), hex_contents],
                GDB_PROT_DONE)

        self.log.debug("Attempted to write memory. Received response: %s" % resp)
        return ret

    def read_memory(self, address, wordsize=4, num_words=1, raw=False):
        """reads memory

        :param address:   Address to write to
        :param wordsize:  the size of a read word (1, 2, 4 or 8)
        :param num_words: the amount of read words
        :param raw:       Whether the read memory should be returned unprocessed
        :return:          The read memory
        """

        num2fmt = {1: 'B', 2: 'H', 4: 'I', 8: 'Q'}

        max_read_size = 0x100
        raw_mem = b''
        for i in range(0, wordsize * num_words, max_read_size):
            to_read = max_read_size if wordsize * num_words > i + max_read_size - 1 else \
                wordsize * num_words % max_read_size
            res, resp = self._sync_request(["-data-read-memory-bytes", str(address + i),
                                            str(to_read)],
                                           GDB_PROT_DONE)

            self.log.debug("Attempted to read memory. Received response: %s" % resp)

            if not res:
                raise Exception("Failed to read memory!")

            # the indirection over the bytearray is needed for legacy python support
            read_mem = bytearray.fromhex(resp['payload']['memory'][0]['contents'])
            raw_mem += bytes(read_mem)

        if raw:
            return raw_mem
        else:
            # Todo: Endianness support
            fmt = '<%d%s' % (num_words, num2fmt[wordsize])
            mem = list(unpack(fmt, raw_mem))

            if num_words == 1:
                return mem[0]
            else:
                return mem

    def read_register(self, reg):
        if reg in self._arch.special_registers:
            return self._read_special_reg_from_name(reg)
        else:
            reg_nr = (
                self._origin.regs._get_nr_from_name(reg)
                if hasattr(self._origin, 'regs')
                else self._arch.registers[reg])
            return self.read_register_from_nr(reg_nr)

    def _read_special_reg_from_name(self, reg):
        """GDB does not return simple values for certain registers,
           such as SSE-registers on x86.
           This function tries to cover those cases by looking up an
           expression to access the register, and the resulting format,
           in the architecture description.
        """

        ret, resp = self._sync_request(
            ["-data-evaluate-expression", "%s" %
                self._arch.special_registers[reg]['gdb_expression']],
             GDB_PROT_DONE)
        fmt = self._arch.special_registers[reg]['format']
        res = parse.parse(fmt, resp['payload']['value'])
        if res is None:
            self.log.critical(
                "Unable to parse content of special register %s" % reg
            )
            raise Exception("Couldn't parse special register")
        return list(res)




    def read_register_from_nr(self, reg_num):
        """Gets the value of a single register

        :param reg_num: number of the register
        :returns:       the value as integer on success, else None
        :todo: Implement function for multiple registers
        """

        ret, resp = self._sync_request(
            ["-data-list-register-values", "x", "%d" % reg_num], GDB_PROT_DONE)

        self.log.debug(
            "Attempted to get register value. Received response: %s" %
            resp)
        return int(resp['payload']['register-values']
                   [0]['value'], 16) if ret else None

    def write_register(self, reg, value):
        """Set one register on the target
        :returns: True on success"""

        if reg in self._arch.special_registers:

            fmt = "{:s}=" \
                  + self._arch.special_registers[reg]['format'].replace(' ','')

            ret, resp = self._sync_request(
                ["-data-evaluate-expression", fmt.format(
                   self._arch.special_registers[reg]['gdb_expression'], *value)
                ], GDB_PROT_DONE
            )
        else:
            ret, resp = self._sync_request(
                ["-gdb-set", "$%s=0x%x" % (reg, value)], GDB_PROT_DONE)

        self.log.debug("Attempted to set register. Received response: %s" % resp)
        return ret

    def step(self):
        """Step one instruction on the target
        :returns: True on success"""
        ret, resp = self._sync_request(
            ["-exec-step-instruction"], GDB_PROT_RUN)

        self.log.debug(
            "Attempted to step on the target. Received response: %s" %
            resp)
        return ret

    def run(self):
        """Starts the execution of the target
        :returns: True on success"""
        ret, resp = self._sync_request(["-exec-run"], GDB_PROT_RUN)

        self.log.debug(
            "Attempted to start execution on the target. Received response: %s" %
            resp)
        return ret

    def cont(self):
        """Continues the execution of the target
        :returns: True on success"""
        ret, resp = self._sync_request(["-exec-continue"], GDB_PROT_RUN)

        self.log.debug(
            "Attempted to continue execution on the target. Received response: %s" %
            resp)
        return ret

    def stop(self):
        """Stops execution of the target
        :returns: True on success"""

        ret, resp = self._sync_request(
            ["-exec-interrupt", "--all"], GDB_PROT_DONE)

        self.log.debug(
            "Attempted to stop execution of the target. Received response: %s" %
            resp)
        return ret

    def set_file(self, elf=''):
        """Load an ELF file
        :returns: True on success"""
        ret, resp = self._sync_request(["-file-exec-and-symbols", elf], GDB_PROT_DONE)

        self.log.debug(
            "Attempted to load elf file. Received response: %s" %
            resp)
        return ret

    def download(self):
        """Download code to target
        :returns: True on success"""
        ret, resp = self._sync_request(["-target-download"], GDB_PROT_DONE, timeout=60)

        self.log.debug(
            "Attempted to download code to target. Received response: %s" %
            resp)
        return ret

    def set_endianness(self, endianness='little'):
        req = ['-gdb-set', 'endian', '%s' % endianness]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)

        self.log.debug("Attempt to set endianness of the target. Received: %s" %
                       resp)
        return ret

    def get_mappings(self):
        self._communicator.start_console_collection()
        req = ['info', 'proc', 'mappings']
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        self._communicator.stop_console_collection()
        self.log.debug("Attempt to read the memory mappings of the target. " +
                       "Received: %s" % resp)
        return ret, self._communicator._console_output

    def console_command(self, cmd, rexpect=GDB_PROT_DONE):
        self._communicator.start_console_collection()
        req = cmd.split()
        ret, resp = self._sync_request(req, rexpect)
        self._communicator.stop_console_collection()
        self.log.debug("Attempt to execute the console command: %s" % cmd)
        return ret, self._communicator._console_output

    def get_symbol(self, symbol):
        self._communicator.start_console_collection()
        req = ['info', 'address', '%s' % symbol]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        self._communicator.stop_console_collection()
        if ret:
            resp = self._communicator._console_output
            regex = re.compile("(0x[0-9a-f]*)[ .]")
            resp = regex.findall(resp)
            if len(resp) == 1:
                resp = int(resp[0], 16)
            else:
                resp = -1
                ret = False
        self.log.debug("Attempt to resolve the symbol %s. " +
                       "Received: %s" % resp)
        return ret, resp

    def set_gdb_variable(self, variable, value):
        req = ['-gdb-set', str(variable), str(value)]
        ret, resp = self._sync_request(req, GDB_PROT_DONE)
        if ret:
            self.log.debug("Successfully set variable %s to %s" %
                           (str(variable), str(value)))
        else:
            self.log.debug("Unable to set variable %s to %s" %
                           (str(variable), str(value)))
        return ret

    def quit(self):
        req = ['-gdb-exit']
        ret = self._sync_request(req, GDB_PROT_DONE)
        return ret



from os.path import abspath
from sys import version_info

from .peripherals.avatar_peripheral import AvatarPeripheral


class MemoryRange(object):
    """
    This class represents a MemoryRange which can be mapped in one of
    Avatar-Targets.

    :ivar address:      The load-address of the memory range
    :ivar size:         The size of the memory range
    :ivar name:         User-defined name for the memory range
    :ivar permissions:  The permisions of the range, represented as textual
                        unix file permission (rwx)
    :ivar file:         A file used for backing the memory range
    :ivar file_offset:  The offset within the file
    :ivar file_bytes:   Bytes of the file to be copied into memory
    :ivar forwarded:    Enable or disable forwarding for this range
    :ivar forwarded_to: List of targets this range should be forwarded to
    :ivar is_symbolic:  Consider this range as symbolic in certain targets
    :ivar is_special:   Whether the range represents special memory which
                        behaves unlike normal memory, e.g. MMIO
    """

    def __init__(self, address, size, name=None, permissions='rwx',
                 file=None, file_offset=None, file_bytes=None, forwarded=False,
                 forwarded_to=None, is_symbolic=False, is_special=False,
                 **kwargs):
        self.address = address
        self.size = size
        self.name = (
            name
            if name is not None else
            "mem_range_0x{:08x}_0x{:08x}".format(address, address+size))
        self.permissions = permissions
        self.file = abspath(file) if file is not None else None
        self.file_offset = file_offset if file_offset is not None else None
        self.file_bytes = file_bytes if file_bytes is not None else None
        self.forwarded = forwarded
        self.is_symbolic = is_symbolic
        self.is_special = is_special
        self.forwarded_to = forwarded_to
        self.__dict__.update(kwargs)


    def dictify(self):
        """
        Returns the memory range as *printable* dictionary
        """
        from .avatar2 import Avatar # we cannot do this import at top-level
        # Assumption: dicts saved in mrs are of primitive types only
        expected_types = (str, bool, int, dict, AvatarPeripheral, Avatar, list)
        if version_info < (3, 0): expected_types += (unicode, )

        tmp_dict = dict(self.__dict__)
        mr_dict = {}
        while tmp_dict != {}:
            k, v = tmp_dict.popitem()
            if v is None or False: continue
            elif k == 'forwarded_to': v = v.name
            # TODO handle emulate
            if not isinstance(v, expected_types):
                raise Exception(
                    "Unsupported type %s for dictifying %s for mem_range at 0x%x"
                    % (type(v), k, self.address))
            if isinstance(v, AvatarPeripheral):
                v = v.__class__.__name__
            if isinstance(v, Avatar):
                v = v.__class__.__name__
            mr_dict[k] = v
        return mr_dict





from threading import Thread
from functools import wraps


class WatchedTypes(object):
    watched_types = [
        'StateTransfer',
        'BreakpointHit',
        'UpdateState',
        'SyscallCatched',
        'RemoteMemoryRead',
        'RemoteMemoryWrite',
        'RemoteInterruptEnter',
        'RemoteInterruptExit',
        'AvatarGetStatus',
        'AddTarget',
        'TargetInit',
        'TargetShutdown',
        'TargetCont',
        'TargetStop',
        'TargetStep',
        'TargetGetSymbol',
        'TargetWriteMemory',
        'TargetReadMemory',
        'TargetRegisterWrite',
        'TargetRegisterRead',
        'TargetSetBreakpoint',
        'TargetSetWatchPoint',
        'TargetRemovebreakpoint',
        'TargetWait',
        'TargetSetFile',
        'TargetDownload',
        'TargetInjectInterrupt'
    ]

    def __init__(self):
        self.watched_types = []
        for type in WatchedTypes.watched_types:
            setattr(self, type, type)
            self.watched_types.append(type)

    def __iter__(self):
        for type in self.watched_types:
            yield type

    def _add(self, type):
        if type not in self.watched_types:
            self.watched_types.append(type)
            setattr(self, type, type)
            return True
        return False


BEFORE = 'before'
AFTER = 'after'


def watch(watched_type):
    """
    Decorator for the watchmen system
    """

    def decorator(func):
        @wraps(func)
        def watchtrigger(self, *args, **kwargs):
            # To avoid circular dependencies, we import here ...
            from .avatar2 import Avatar
            from .targets.target import Target

            cb_kwargs = dict(kwargs)
            if isinstance(self, Avatar):
                avatar = self
            elif isinstance(self, Target):
                avatar = self.avatar
                cb_kwargs['watched_target'] = self

            avatar.watchmen.t(watched_type, BEFORE, *args, **cb_kwargs)
            ret = func(self, *args, **kwargs)
            cb_kwargs.update({'watched_return': ret})
            cb_ret = avatar.watchmen.t(watched_type, AFTER, *args, **cb_kwargs)

            return ret if cb_ret is None else cb_ret

        return watchtrigger

    return decorator


class AsyncReaction(Thread):
    def __init__(self, avatar, callback, *args, **kwargs):
        super(AsyncReaction, self).__init__()
        self.avatar = avatar
        self.callback = callback
        self.args = args
        self.kwargs = kwargs

    def run(self):
        self.callback(self.avatar, *self.args, **self.kwargs)


class WatchedEvent(object):
    # noinspection PyUnusedLocal
    def __init__(self, watch_type, when, callback, is_async,
                 overwrite_return = False, *args, **kwargs):
        self._callback = callback
        self.type = watch_type
        self.when = when
        self.is_async = is_async
        self.overwrite_return = overwrite_return
        self.watchmen_args = args
        self.watchmen_kwargs = kwargs

    def react(self, avatar, *args, **kwargs):
        if self._callback is None:
            raise Exception("No callback defined for watchmen of type %s" %
                            self.type)
        else:
            args += self.watchmen_args
            # WARNING: This could overwrite original kwargs,
            #          which may be desired in some cases
            kwargs.update(self.watchmen_kwargs)
            if self.is_async:
                thread = AsyncReaction(avatar, self._callback, *args, **kwargs)
                self.daemon = True
                thread.start()
            else:
                ret = self._callback(avatar, *args, **kwargs)
                if self.overwrite_return:
                    return ret
                


class Watchmen(object):
    """
    """

    def __init__(self, avatar):
        self._watched_events = {}
        self._avatar = avatar
        self.watched_types = WatchedTypes()

        for e in self.watched_types:
            self._watched_events[e] = []

    def add_watch_types(self, watched_types):
        for type in watched_types:
            if self.watched_types._add(type):
                self._watched_events[type] = []

    def add_watchman(self, watch_type, when=BEFORE, callback=None, is_async=False, overwrite_return=False, *args, **kwargs):

        if watch_type not in self.watched_types:
            raise Exception("Requested event_type does not exist")
        if when not in (BEFORE, AFTER):
            raise Exception("Watchman has to be invoked \'before\' or \'after\'!")
        if when == BEFORE and overwrite_return == True:
            self._avatar.log.warning("Overite return enabled for watchman BEFORE event %s. Discarding." % watch_type)
            overwrite_return = False
        w = WatchedEvent(watch_type, when, callback, is_async,
                         overwrite_return=overwrite_return,
                         *args, **kwargs)
        overwriting_cbs = [x.overwrite_return for x in
                           self._watched_events[watch_type]
                           if x.overwrite_return]

        if len(overwriting_cbs) > 1:
            self._avatar.log.warning("More than one watchman can modify the return value for the event. Watch type: %s" % watch_type)
        self._watched_events[watch_type].append(w)
        return w

    add = add_watchman

    def remove_watchman(self, watch_type, watchman):
        if watch_type not in self.watched_types:
            raise Exception("Requested event_type does not exist")
        self._watched_events[watch_type].remove(watchman)

    def trigger(self, watch_type, when, *args, **kwargs):
        ret = None
        for watchman in self._watched_events[watch_type]:
            if watchman.when == when:
                if watchman.overwrite_return:
                    ret = watchman.react(self._avatar, *args, **kwargs)
                    kwargs.update({'watched_return': ret})
                else:
                    watchman.react(self._avatar, *args, **kwargs)
        return ret

    t = trigger



import sys
import pylink
from time import sleep
from threading import Thread, Event, Condition
import logging
import re

if sys.version_info < (3, 0):
    import Queue as queue
    # __class__ = instance.__class__
else:
    import queue

from avatar2.archs.arm import ARM
from avatar2.targets import TargetStates
from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage


class JLinkProtocol(Thread):
    """Main class for the JLink bprotocol, via pylink-square
    :ivar serial: The serial number of the JLink to connect to
    :ivar device: The JLink device name for the target
    :ivar avatar:   the avatar object
    :ivar origin:   the target utilizing this protocol
    """

    def __init__(self, serial="12345678", device="ARM7", avatar=None, origin=None):
        self._shutdown = Event()
        self.avatar = avatar
        self.origin = origin
        self.jlink = pylink.JLink()
        self.jlink.open(serial)
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)
        Thread.__init__(self)
        self.connect(device=device)

    def __del__(self):
        self.shutdown()

    def connect(self, device="ARM7"):
        # Todo add a time out here
        while True:
            try:
                self.jlink.connect(device, verbose=True)
                self.jlink.ir_len()
                break
            except pylink.errors.JLinkException:
                self.log.info("Connection failed, trying again...")
                sleep(0.25)
        self.log.info("Connected to JLink target")
        self.start()
        return True

    def reset(self, halt=True):
        self.log.info("Resetting target")
        return self.jlink.reset(halt=halt)

    def shutdown(self):
        self._shutdown.set()
        
    def update_target_regs(self):
        """
        This function will try to update the TargetRegs based on the list of
        registers known to gdb.
        """
        regs = {}
        for idx in self.jlink.register_list():
            name = self.jlink.register_name(idx)
            regs[name] = idx

        if hasattr(self.origin, 'regs'):
            self.origin.regs._update(regs)

    def run(self):
        # Target state management thread
        # This thread needs to poll for the halted state
        # of the target
        # JLink is lame and doesn't let you do this asynch
        # Also, not all targets produce a "moe" (Mode of Entry) 
        # so we have to actually do that here.
        try:
            while not self._shutdown.is_set():
                is_halted = self.jlink.halted()
                if is_halted and self.origin.state == TargetStates.RUNNING:
                    # We just halted
                    # But did we hit a BP?
                    self.log.debug("JLink Target is halting...")
                    avatar_msg = UpdateStateMessage(self.origin, TargetStates.STOPPED)
                    self.avatar.fast_queue.put(avatar_msg)
                    self.origin.wait()
                    self.log.debug("JLink target has halted")
                    pc = self.get_pc()
                    if self.jlink.breakpoint_find(pc):
                        self.log.debug("JLink Target hit breakpoint %d" % self.jlink.breakpoint_find(pc))
                        avatar_msg = BreakpointHitMessage(self.origin, self.jlink.breakpoint_find(pc), pc)
                        self.avatar.queue.put(avatar_msg)

                elif not is_halted and self.origin.state == TargetStates.STOPPED:
                    self.log.debug("About to resume target.")
                    avatar_msg = UpdateStateMessage(self.origin, TargetStates.RUNNING)
                    self.avatar.fast_queue.put(avatar_msg)
                    while self.origin.state != TargetStates.RUNNING:
                        pass
                    self.log.debug("JLink target has resumed")
        except:
            self.log.exception("JLink target errored")
        finally:
            self.log.info("JLink target exiting")
            self.jlink.close()

    def set_breakpoint(self, line,
                       hardware=False,
                       temporary=False,
                       regex=False,
                       condition=None,
                       ignore_count=0,
                       thread=0,
                       pending=False):
        """Inserts a breakpoint

        :param bool hardware: Hardware breakpoint
        :param bool temporary:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with specified condition
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        :returns:             The number of the breakpoint
        """
        # TODO: Hw/Sw breakpoint control
        self.log.info("Setting breakpoint at %#08x" % line)
        ret = self.jlink.breakpoint_set(line)
        self.log.info("Got BP ID %d" % ret)
        return ret

    def set_watchpoint(self, variable, write=True, read=False):
        return self.jlink.watchpoint_set(variable, write=write, read=read)

    def remove_breakpoint(self, bkpt):
        """Deletes a breakpoint"""
        # TODO: Check this
        return self.jlink.breakpoint_clear(bkpt)

    def write_memory(self, address, wordsize, val, num_words=1, raw=False):
        """Writes memory

        :param address:   Address to write to
        :param wordsize:  the size of the write (1, 2, 4 or 8)
        :param val:       the written value
        :type val:        int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """
        if raw:
            new_val = []
            if not len(val):
                raise ValueError("val had zero length")
            new_val = [ord(v) for v in val]
            val = new_val
        try:
            self.jlink.memory_write(address, contents)
            return True
        except pylink.JLinkException:
            return False

    def read_memory(self, address, wordsize=4, num_words=1, raw=False):
        """reads memory

        :param address:   Address to write to
        :param wordsize:  the size of a read word (1, 2, 4 or 8) 
        :param num_words: the amount of read words
        :param raw:       Whether the read memory should be returned unprocessed
        :return:          The read memory
        """

        ret = self.jlink.memory_read(address, num_units=num_words, nbits=wordsize)

        if raw:
            raw_mem = "".join([newint.to_bytes(i, length=int(math.ceil(i.bit_length() / 8.0))) for i in ret])
            return raw_mem
        return ret

    def read_register(self, reg):
        the_reg = tolower(reg)
        the_idx = -1
        for idx in self.jlink.register_list():
            if the_reg == self.jlink.register_name(idx):
                the_idx = idx
                break
        return self.register_read(the_idx)

    def get_pc(self):
        # Get PC a shitty way
        for idx in self.jlink.register_list():
            if "PC" in self.jlink.register_name(idx):
                return self.jlink.register_read(idx)

    def write_register(self, reg, val):
        """Set one register on the target
        :returns: True on success"""
        the_reg = tolower(reg)
        the_idx = -1
        for idx in self.jlink.register_list():
            if the_reg == self.jlink.register_name(idx):
                the_idx = idx
                break
        return self.jlink.register_write(the_idx, val)
        
    def step(self):
        """Step one instruction on the target
        :returns: True on success"""
        return self.jlink.step()

    def cont(self):
        """Continues the execution of the target
        :returns: True on success"""
        self.log.info("Resuming target...")
        return self.jlink.restart()

    def stop(self):
        """Stops execution of the target
        :returns: True on success"""
        self.log.info("Stopping target...")
        return self.jlink.halt()

    def set_endianness(self, endianness='little'):
        if 'little' in endianness:
            self.jlink.set_little_endian()
        elif "big" in endianness:
            self.jlink.set_big_endian()

		



import sys
from threading import Thread, Event, Condition
from struct import pack, unpack
from codecs import encode
import logging
import os
import re
from bitstring import BitStream, ReadError
from binascii import unhexlify
import pygdbmi.gdbcontroller
from .openocd import OpenOCDProtocol

if sys.version_info < (3, 0):
    import Queue as queue
    # __class__ = instance.__class__
else:
    import queue

from avatar2.archs.arm import ARM
from avatar2.targets import TargetStates
from avatar2.message import AvatarMessage, UpdateStateMessage, \
    BreakpointHitMessage, RemoteInterruptEnterMessage
from avatar2.protocols.openocd import OpenOCDProtocol

logger = logging.getLogger(__name__)

# ARM System Control Block
SCB_CPUID = 0xe000ed00  # What is it
SCB_STIR = 0xe000ef00  # Send interrupts here
SCB_VTOR = 0xe000ed08  # Vector Table offset register

# NVIC stuff
NVIC_ISER0 = 0xe000e100

# CoreSight Constant Addresses
RCC_APB2ENR = 0x40021018
AFIO_MAPR = 0x40010004
DBGMCU_CR = 0xe0042004
COREDEBUG_DEMCR = 0xe000edfc
TPI_ACPR = 0xe0040010
TPI_SPPR = 0xe00400f0
TPI_FFCR = 0xe0040304
DWT_CTRL = 0xe0001000
ITM_LAR = 0xe0000fb0
ITM_TCR = 0xe0000e80
ITM_TER = 0xe0000e00
ETM_LAR = 0xe0041fb0
ETM_CR = 0xe0041000
ETM_TRACEIDR = 0xe0041200
ETM_TECR1 = 0xe0041024
ETM_FFRR = 0xe0041028
ETM_FFLR = 0xe004102c


class CoreSightProtocol(Thread):
    def __init__(self, avatar, origin):
        self.avatar = avatar
        self._avatar_queue = avatar.queue
        self._avatar_fast_queue = avatar.fast_queue
        self._origin = origin
        self.trace_queue = None
        self.trace_buffer = BitStream()
        self._close = Event()
        self._closed = Event()
        self._close.clear()
        self._closed.clear()
        self._sync_responses_cv = Condition()
        self._last_exec_token = 0
        # logger = logging.getLogger('%s.%s' %
        #                              (origin.log.name, self.__class__.__name__)
        #                              ) if origin else \
        #     logging.getLogger(self.__class__.__name__)
        self._monitor_stub_base = None
        self._monitor_stub_isr = None
        self._monitor_stub_loop = None
        self._monitor_stub_writeme = None
        Thread.__init__(self)
        self.daemon = True

    def __del__(self):
        self.shutdown()

    def inject_interrupt(self, interrupt_number, cpu_number=0):
        # Set an interrupt using the STIR
        self._origin.write_memory(SCB_STIR, 4, interrupt_number)

    def enable_interrupt(self, interrupt_number):
        """
        Enables an interrupt (e.g., in the NIVC)
        :param interrupt_number:
        :return:
        """
        assert (0 < interrupt_number < 256)
        iser_num = interrupt_number >> 5
        iser_addr = NVIC_ISER0 + (iser_num * 4)
        # iser_off = interrupt_number % 32
        # iser_val = self._origin.read_memory(iser_addr, 4)
        iser_val = ((1 << interrupt_number) & 0x1F)
        # iser_val |= 0x1 << iser_off
        self._origin.write_memory(iser_addr, 4, iser_val)

    def get_vtor(self):
        return self._origin.read_memory(SCB_VTOR, 4)

    def get_ivt_addr(self):
        if getattr(self._origin, 'ivt_address', None) is not None:
            return self._origin.ivt_address
        else:
            return self.get_vtor()

    def set_vtor(self, addr):
        return self._origin.write_memory(SCB_VTOR, 4, addr)

    def get_isr(self, interrupt_num):
        return self._origin.read_memory(
            self.get_ivt_addr() + (interrupt_num * 4), 4)

    def set_isr(self, interrupt_num, addr):
        return self._origin.write_memory(
            self.get_ivt_addr() + (interrupt_num * 4), 4, addr)

    def cpuid(self):
        c = self._origin.read_memory(SCB_CPUID, 4, 1)
        print("CPUID: %#08x" % c)
        if (0x412fc230 & 0x000f0000) >> 16 == 0xf:
            print("Found ARM Cortex CPUID")
        else:
            return
        impl = (c >> 24)
        vari = (c & 0x00f00000) >> 20
        part = (c & 0x0000fff0) >> 4
        rev = (c & 0x0000000f)
        print("Implementer %#08x, Variant %#08x, Part %#08x, Rev %#08x" % (
            impl, vari, part, rev))

    def shutdown(self):
        if self.is_alive() is True:
            self.stop()

    def connect(self):
        if not isinstance(self._origin.protocols.monitor, OpenOCDProtocol):
            raise Exception(("CoreSightProtocol requires OpenOCDProtocol ")
                            ("to be present."))

    def has_bits_to_read(self, b, n):
        return b.len - b.pos > n

    def enable_interrupts(self, use_tcl_tracing=False, use_stub=True):
        try:
            logger.info("Starting CoreSight Protocol")
            if not isinstance(self._origin.protocols.monitor, OpenOCDProtocol):
                raise Exception(
                    "CoreSightProtocol requires OpenOCDProtocol to be present.")
            openocd = self._origin.protocols.monitor
            logger.debug("Resetting target")
            openocd.reset()

            if use_tcl_tracing:
                # Enable TCL tracing
                if not openocd.trace_enabled.is_set():
                    openocd.enable_trace()
                    if not openocd.trace_enabled.is_set():
                        logger.error(
                            "Can't get trace events without tcl_trace! aborting...")
                        return False
                self.trace_queue = openocd.trace_queue
                # Enable the TPIO output to the FIFO
                logger.debug("Enabling TPIU output events")
                openocd.execute_command(
                    'tpiu config internal - uart off 32000000')
                # Enable the DWT to get interrupts
                logger.debug("Enabling exceptions in DWT")
                openocd.execute_command(
                    "setbits $COREDEBUG_DEMCR 0x1000000")  # Enable access to trace regs - set TRCENA to 1
                openocd.execute_command(
                    "mww $DWT_CTRL 0x40010000")  # exc trace only
                logger.debug("Enabling ITM passthrough of DWT events")
                # Enable the ITM to pass DWT output to the TPIU
                openocd.execute_command("mww $ITM_LAR 0xC5ACCE55")
                openocd.execute_command(
                    "mww $ITM_TCR 0x0000000d")  # TraceBusID 1, enable dwt/itm/sync
                openocd.execute_command(
                    "mww $ITM_TER 0xffffffff")  # Enable all stimulus ports
                # Run our little daemon thingy
                logger.debug("Starting interrupt handling thread")
                self.daemon = True
                self.start()

            elif use_stub:
                self.inject_monitor_stub()
        except:
            logger.exception("Error starting Coresight")

    """
    What this does:
    Hang in a loop at `loop`
    When an interrupt comes, go to `stub`
    At `stub`, load `writeme`, if it's not zero, reset it, and jump to the written value.
    This lets us inject exc_return values into the running program
    """
    # MONITOR_STUB = """
    # loop: b loop
    # nop
    # mov r2, pc
    # ldr r1, [r2, #16]
    # stub:
    # ldr r0, [r2, #12]
    # cmp r1, r0
    # beq stub
    # str r1, [r2, #12]
    # bx r0
    # nop
    # writeme: .word 0xffffffff
    # loadme: .word 0xffffffff
    # """
    # str r2, [r1]

    MONITOR_STUB = """
    dcscr:   .word 0xe000edf0
    haltme:  .word 0xA05F0003
    writeme: .word 0x00000000
    init:
    ldr r1, =dcscr
    ldr r2, =haltme
    ldr r3, =writeme
    ldr r1, [r1]
    ldr r2, [r2]
    loop: b loop
    stub: 
    nop
    intloop:
    ldr r4, [r3]
    cmp r4, #0
    beq intloop
    ldr r4, #0
    str r4, [r3]
    bx lr
    """

    def get_user_pc(self):
        """
        Return the "user PC", that is, the PC at the time an interrupt occurred.
        Returns None if we're not in an interrupt right now.

        :return:
        """
        if self.get_current_isr_num() > 0:
            sp = self._origin.get_register('sp')
            val = self._origin.read_memory(sp - 24)
            return val
        return None

    def get_current_isr_num(self):
        """
        If we're in an interrupt, return the current ISR number that we're in.

        :return:
        """
        # The bottom 8 bits of xPSR
        xpsr = self._origin.read_register("xPSR")
        xpsr &= 0xff
        return xpsr

    def inject_monitor_stub(self, addr=0x20001234, vtor=0x20002000):
        """
        Injects a safe monitoring stub.
        This has the following effects:
        0. Pivot the VTOR to someplace sane
        1. Insert an infinite loop at addr
        2. Set the PC to addr
        3. set up logic for the injection of interrupt returns.
           Write to return_code_register to trigger an IRET
        4.
        :return:
        """
        logger.info(
            "Injecting monitor stub. (IVT: %s, %s, %s)"
            ")" % (self.get_ivt_addr(), self.get_vtor(), vtor))

        self._monitor_stub_base = addr
        self._monitor_stub_loop = addr + 12
        self._monitor_stub_isr = addr + 25
        self._monitor_stub_writeme = addr + 8
        # Pivot VTOR, if needed
        # On CM0, you can't, so don't.
        if getattr(self._origin, 'ivt_address', None) is None:
            if self.get_vtor() == 0:
                self.set_vtor(vtor)
        # Sometimes, we need to gain access to the IVT (make it writable). Do that here.
        if getattr(self._origin, 'ivt_unlock', None) is not None:
            unlock_addr, unlock_val = self._origin.ivt_unlock
            self._origin.write_memory(unlock_addr, 4, unlock_val)
        # put the stub
        self._origin.inject_asm(self.MONITOR_STUB, self._monitor_stub_base)
        # wreck the IVT
        # ...oh but don't wipe out the 0'th position.
        for x in range(1, 254):
            self.set_isr(x, self._monitor_stub_isr)
        if self._origin.state != TargetStates.STOPPED:
            logger.warning(
                "Not setting PC to the monitor stub; Target not stopped")
        else:
            self._origin.regs.pc = self._monitor_stub_loop

    def inject_exc_return(self, exc_return):
        if not self._monitor_stub_base:
            logger.error(
                "You need to inject the monitor stub before you can inject exc_returns")
            return False
        # We can just BX LR for now.
        return self._origin.write_memory(self._monitor_stub_writeme, 4, 1)

    def dispatch_exception_packet(self, packet):
        int_num = ((ord(packet[1]) & 0x01) << 8) | ord(packet[0])
        transition_type = (ord(packet[1]) & 0x30) >> 4

        msg = RemoteInterruptEnterMessage(self._origin, transition_type,
                                          int_num)
        self._avatar_fast_queue.put(msg)

    def run(self):
        DWT_PKTSIZE_BITS = 24
        trace_re = re.compile("type target_trace data ([0-9a-f]+)")
        logger.debug("Starting interrupt thread")
        try:
            while not self._close.is_set():
                if self._close.is_set():
                    break
                # OpenOCD gives us target_trace events packed with many, many packets.
                # Get them out, then do them packet-at-a-time
                if not self.has_bits_to_read(self.trace_buffer,
                                             DWT_PKTSIZE_BITS):
                    # get some more data
                    if self.trace_queue.empty():
                        # make sure we can see the shutdown flag
                        continue
                    new_data = self.trace_queue.get()
                    m = trace_re.match(new_data)
                    if m:
                        self.trace_buffer.append("0x" + m.group(1))
                    else:
                        raise ValueError(
                            "Got a really weird trace packet " + new_data)
                if not self.has_bits_to_read(self.trace_buffer,
                                             DWT_PKTSIZE_BITS):
                    continue
                try:
                    pkt = self.trace_buffer.peek(DWT_PKTSIZE_BITS).bytes
                except ReadError:
                    logger.error("Fuck you length is " + repr(
                        len(self.trace_buffer)) + " " + repr(DWT_PKTSIZE_BITS))
                if ord(pkt[0]) == 0x0E:  # exception packets
                    pkt = pkt[1:]
                    self.dispatch_exception_packet(pkt)
                    # eat the bytes
                    self.trace_buffer.read(DWT_PKTSIZE_BITS)
                # the first byte didn't match, rotate it out
                else:
                    self.trace_buffer.read(8)
        except:
            logger.exception("Error processing trace")
        self._closed.set()
        logger.debug("Interrupt thread exiting...")

    def stop(self):
        """Stops the listening thread. Useful for teardown of the target"""
        self._close.set()
        self._closed.wait()



import sys

if sys.version_info < (3, 0):
    from Queue import PriorityQueue
else:
    from queue import PriorityQueue
import time
import os
from avatar2.targets import Target
from avatar2.protocols.openocd import OpenOCDProtocol
from avatar2.protocols.gdb import GDBProtocol
from avatar2.watchmen import watch


class OpenOCDTarget(Target):
    def __init__(self, avatar, executable=None,
                 openocd_script=None, additional_args=None,
                 tcl_port=6666,
                 gdb_executable=None, gdb_additional_args=None, gdb_port=3333,
                 **kwargs
                 ):

        if openocd_script and not os.path.exists(openocd_script):
            raise ValueError("OpenOCD script %s does not exist!" % openocd_script)
        super(OpenOCDTarget, self).__init__(avatar, **kwargs)

        self.executable = (executable if executable is not None
                           else self._arch.get_oocd_executable())
        self.avatar = avatar
        self.openocd_script = openocd_script
        self.additional_args = additional_args if additional_args else []

        self.gdb_executable = (gdb_executable if gdb_executable is not None
                               else self._arch.get_gdb_executable())
        self.tcl_port = tcl_port
        self.gdb_additional_args = gdb_additional_args if gdb_additional_args else []
        self.gdb_port = gdb_port

    @watch("TargetInit")
    def init(self):
        openocd = OpenOCDProtocol(self.avatar, self, self.openocd_script,
                                  openocd_executable=self.executable,
                                  additional_args=self.additional_args,
                                  tcl_port=self.tcl_port,
                                  gdb_port=self.gdb_port,
                                  output_directory=self.avatar.output_directory)
        time.sleep(.1)  # give openocd time to start. Find a better solution?
        self.log.debug("Connecting to OpenOCD telnet port")
        ocd_connected = openocd.connect()

        gdb = GDBProtocol(gdb_executable=self.gdb_executable,
                          arch=self._arch,
                          additional_args=self.gdb_additional_args,
                          avatar=self.avatar, origin=self)
        self.log.debug("Connecting to OpenOCD GDB port")
        gdb_connected = gdb.remote_connect(port=self.gdb_port)
        script_has_reset = False
        if self.openocd_script:
            with open(self.openocd_script) as f:
                script = f.read()
            if "reset halt" in script:
                self.log.debug("Not resetting target, script may have done it already")
                script_has_reset = True
        if ocd_connected:
            self.log.info("Successfully connected to OpenOCD target!")
        else:
            self.log.error("Failed to connect to OpenOCD target!")
        if ocd_connected and not script_has_reset:
            self.log.debug("Resetting target...")
            openocd.reset()

        self.protocols.set_all(gdb)
        self.protocols.monitor = openocd
        self.wait()



import logging
import time
from sys import version_info
from functools import wraps
from threading import Event
from types import MethodType

from enum import IntEnum

from ..watchmen import watch


def action_valid_decorator_factory(state, protocol):
    """
    This decorator factory  is used to generate decorators which  verify that
    requested actions on a target, such as step(), stop(), read_register(), 
    write_register() and so on are actually executable.

    :param state: A mask specifying the required state of the Target
    :type state:  An entry of the Enum TargetStates
    :param protocol: The protocol required to execute the action.
    :type protocol: str
    """

    def decorator(func):
        @wraps(func)
        def check(self, *args, **kwargs):
            if getattr(self.protocols, protocol) is None:
                raise Exception(
                    "%s() requested but %s is undefined." %
                    (func.__name__, protocol))
            if not self.state.value & state.value:
                raise Exception("%s() requested but Target is %s" %
                                (func.__name__, TargetStates(self.state).name))
            return func(self, *args, **kwargs)

        return check

    return decorator

def synchronize_state(*states, **kwargs):
    """
    This decorator can be used to make sure that the target executed a desired
    set of state transitions in an particular order.
    This is useful, when the user explicitly requests the target to change
    it's state and need an update notification on the transition itself.
    Internally, this works by creating an event and using a watchmen to check
    whether it was triggered.
    :param *states: The desired states of the target
    :param transition_optional: Also allow to return if the target is already
                                in the desired states, even if the transition
                                didn't happen
    """
    def decorator(func):
        @wraps(func)
        def state_synchronizer(self, *args, **kwargs):
            state = states[-1]
            transition_optional = kwargs.get('transition_optional', False)

            blocking = kwargs.get('blocking', True)
            avatar = self.avatar
            if blocking is True:
                state_reached = Event()

                def state_synchronize_cb(avatar, message, *args, **kwargs):
                    if message.origin == self:
                        if message.state == state:
                            state_reached.set()
                            avatar.watchmen.remove_watchman('UpdateState', w)
                        elif message.state == TargetStates.EXITED:
                            raise Exception("Target %s exited" % self.name)

                w = avatar.watchmen.add('UpdateState', when='after',
                                        callback=state_synchronize_cb)
            if len(states) == 1:
                ret = func(self, *args, **kwargs)
            else:
                ret = synchronize_state(*states[:-1])(func)(self, *args, **kwargs)
            if blocking is True:
                if not (transition_optional == True and self.state == state):
                    state_reached.wait()
            return ret

        return state_synchronizer

    return decorator


class TargetStates(IntEnum):
    """
    A simple Enum for the different states a target can be in.
    """
    CREATED = 0x1
    INITIALIZED = 0x2
    STOPPED = 0x4
    RUNNING = 0x8
    SYNCING = 0x10
    EXITED = 0x20
    NOT_RUNNING = INITIALIZED | STOPPED
    BREAKPOINT = SYNCING | STOPPED

class TargetRegs(object):
    def __init__(self, target, register_dict):
        self._target = target
        self.__dict__.update(register_dict)

    def __getattribute__(self, name):
        if name == '_get_names' or name == '__dict__':
            return super(TargetRegs, self).__getattribute__(name)
        elif name in self._get_names():
            return self._target.read_register(name)
        else:
            return super(TargetRegs, self).__getattribute__(name)

    def __setattr__(self, name, value):
        if name == '_target':
            return super(TargetRegs, self).__setattr__(name, value)
        elif name in self._get_names():
            return self._target.write_register(name, value)
        else:
            return super(TargetRegs, self).__setattr__(name, value)

    def _update(self, reg_dict):
        self.__dict__.update(reg_dict)

    def _get_nr_from_name(self, reg_name):
        return self.__dict__[reg_name]

    def _get_names(self):
        names = set(self.__dict__) ^ set(['_target'])
        return names

class TargetProtocolStore(object):
    """This class stores the various protocols associated to one target"""

    DEFAULT_PROTOCOLS = ['memory', 'registers', 'execution']

    def __init__(self, additional_protocols=None):
        self.protocols = set(TargetProtocolStore.DEFAULT_PROTOCOLS)
        self.protocols |= additional_protocols if additional_protocols else set()
        self.unique_protocols = {} #Stores protocol references and their count
        for p in self.protocols:
            setattr(self, p, None)

    def set_all(self, instance, only_defaults=False):
        """
        Sets an instantiated protocol object for either all protocols in this
        store, or only the default ones
        :param instance: the protocol instance
        """
        protocols = (TargetProtocolStore.DEFAULT_PROTOCOLS if only_defaults
                     else self.protocols
                    )
        for p in protocols:
            setattr(self, p, instance)

    def shutdown(self):
        """Shutsdown all the associated protocols"""
        for p in self.protocols:
            #print("Unloading %s" % str(p))
            setattr(self, p, None)

    def __setattr__(self, name, value):
        if name == 'protocols' or name == 'unique_protocols':
            return super(TargetProtocolStore, self).__setattr__(name, value)

        # Check whether the protocol is already an attribute
        if hasattr(self, name) is False:
            self.protocols.add(name)
            saved_val = None
        else:
            saved_val = getattr(self, name)

        if value is not None and self.unique_protocols.get(value, None) is None:
            self.unique_protocols[value] = 0

        if value is None and saved_val is not None:
            self.unique_protocols[saved_val] -= 1
        elif value is not None and saved_val is None:
            self.unique_protocols[value] += 1
        elif value is not None and saved_val is not None:
            self.unique_protocols[value] += 1
            self.unique_protocols[saved_val] -= 1
     
        # if there is no reference left, let's shut the prot down
        if saved_val is not None and self.unique_protocols[saved_val] == 0:
            getattr(self, name).shutdown()

        return super(TargetProtocolStore, self).__setattr__(name, value)



class Target(object):
    """The Target object is one of Avatars core concept, as Avatar orchestrate 
    different targets.
    While the generic target has no implementation, it provides an insight over 
    all the functions a Target MUST implement
    """

    def __init__(self, avatar, name=None):  # type: ('Avatar', str) -> None
        """
        Creates a new instance of a Target.
        :param avatar: The avatar instance this target should be orchestrated by
        :param name: The name of this target, mainly for logging. This is optional and will be autogenerated otherwise.
        """
        super(Target, self).__init__()
        self.state = TargetStates.CREATED

        self.avatar = avatar
        self.name = name if name else self._get_unique_name()

        self.status = {}
        self._arch = avatar.arch
        self.protocols = TargetProtocolStore()

        self.state = TargetStates.CREATED

        self.log = logging.getLogger('%s.targets.%s' % (avatar.log.name, self.name))
        log_file = logging.FileHandler('%s/%s.log' % (avatar.output_directory, self.name))
        formatter = logging.Formatter('%(asctime)s | %(name)s.%(levelname)s | %(message)s')
        log_file.setFormatter(formatter)
        self.log.addHandler(log_file)

        self.regs = TargetRegs(self, self._arch.registers)

    def dictify(self):
        """
        Returns the memory range as *printable* dictionary for the config
        """

        ignore = ['state', 'status', 'regs', 'protocols', 'log', 'avatar']
        ignored_types = (MethodType)
        expected_types = (str, bool, int, list) 
        if version_info < (3, 0): expected_types += (unicode, )

        t_dict = {'type': self.__class__.__name__,
                  'module': self.__module__}

        for k, v in self.__dict__.items():
            if k in ignore: continue
            if k.startswith('_'): continue
            if v is None: continue
            if isinstance(v, ignored_types): continue
            if not isinstance(v, expected_types):
                raise Exception(
                    "Unsupported type %s for dictifying %s for target %s" %
                    (type(v), k, self.name))
            t_dict[k] = v
        return t_dict



    @watch('TargetInit')
    def init(self):
        """
        Initializes the target to start the analyses
        """
        pass

    @watch('TargetShutdown')
    def shutdown(self):
        """
        Shutdowns the target
        """
        self.protocols.shutdown()

    @watch('TargetCont')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
    @synchronize_state(TargetStates.RUNNING)
    def cont(self, blocking=True):
        """
        Continues the execution of the target
        :param blocking: if True, block until the target is RUNNING
        """
        return self.protocols.execution.cont()


    @watch('TargetStop')
    @action_valid_decorator_factory(TargetStates.RUNNING, 'execution')
    @synchronize_state(TargetStates.STOPPED, transition_optional=True)
    def stop(self, blocking=True):
        return self.protocols.execution.stop()

    @watch('TargetStep')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
    @synchronize_state(TargetStates.RUNNING, TargetStates.STOPPED)
    def step(self, blocking=True):
        """
        Steps one instruction.
        :param blocking: if True, block until the target is STOPPED again
        """
        return self.protocols.execution.step()

    @watch('TargetSetFile')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
    def set_file(self, elf):
        """
        Load an ELF file

        :param elf: ELF file to load
        :returns: True on success else False
        """
        if not hasattr(self.protocols.execution, 'set_file'):
            self.log.error('Protocol "' + type(self.protocols.execution).__name__ + '" does not support "set_file"')
            return False

        return self.protocols.execution.set_file(elf)

    @watch('TargetDownload')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
    def download(self):
        """
        Download the loaded code to the Target

        :returns: True on success else False
        """
        if not hasattr(self.protocols.execution, 'download'):
            self.log.error('Protocol "' + type(self.protocols.execution).__name__ + '" does not support "download"')
            return False

        return self.protocols.execution.download()

    @watch('TargetGetSymbol')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
    def get_symbol(self, symbol):
        """
        Get the address of a symbol

        :param symbol:    The name of a symbol whose address is wanted
        :returns:         (True, Address) on success else False
        """
        return self.protocols.memory.get_symbol(symbol)

    @watch('TargetWriteMemory')
    #@action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
    def write_memory(self, address, size, value, num_words=1, raw=False):
        """
        Writing to memory of the target

        :param address:   The address from where the memory-write should 
                          start
        :param size:      The size of the memory write 
        :param value:     The actual value written to memory
        :type val:        int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """
        try:
            target_range = self.avatar.get_memory_range(address)
        except Exception as e:
            self.log.warn("Performing write on undefined range at 0x%x" % address)
            target_range = None
        if target_range is not None and target_range.forwarded is True and \
        target_range.forwarded_to != self:
            return target_range.forwarded_to.write_memory(address, size, value,
                                                          num_words, raw)
        
        return self.protocols.memory.write_memory(address, size, value,
                                                  num_words, raw)

    @watch('TargetReadMemory')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'memory')
    def read_memory(self, address, size, num_words=1, raw=False):
        """
        Reading from memory of the target

        :param address:     The address to read from 
        :param size:        The size of a read word
        :param num_words:   The amount of words to read (default: 1)
        :param raw:         Whether the read memory is returned unprocessed
        :return:          The read memory
        """
        try:
            target_range = self.avatar.get_memory_range(address)
        except Exception as e:
            self.log.warn("Performing write on undefined range at 0x%x" % address)
            target_range = None
        if target_range is not None and target_range.forwarded is True and \
        target_range.forwarded_to != self:
            return target_range.forwarded_to.read_memory(address, size,
                                                         num_words, raw)
        
        return self.protocols.memory.read_memory(address, size, num_words, raw)

    @watch('TargetRegisterWrite')
    #@action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
    def write_register(self, register, value):
        """
        Writing a register to the target

        :param register:     The name of the register
        :param value:        The actual value written to the register
        """
        return self.protocols.registers.write_register(register, value)

    @watch('TargetRegisterRead')
    #@action_valid_decorator_factory(TargetStates.STOPPED, 'registers')
    def read_register(self, register):
        """
        Reading a register from the target

        :param register:     The name of the register
        :return:             The actual value read from the register
        """
        return self.protocols.registers.read_register(register)

    @watch('TargetSetBreakpoint')
    @action_valid_decorator_factory(TargetStates.NOT_RUNNING, 'execution')
    def set_breakpoint(self, line, hardware=False, temporary=False, regex=False,
                       condition=None, ignore_count=0, thread=0, **kwargs):
        """Inserts a breakpoint

        :param bool hardware: Hardware breakpoint
        :param bool tempory:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with the condition
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        """
        return self.protocols.execution.set_breakpoint(line, hardware=hardware,
                                                  temporary=temporary,
                                                  regex=regex,
                                                  condition=condition,
                                                  ignore_count=ignore_count,
                                                  thread=thread, **kwargs)

    @watch('TargetSetWatchPoint')
    @action_valid_decorator_factory(TargetStates.NOT_RUNNING, 'execution')
    def set_watchpoint(self, variable, write=True, read=False):
        """Inserts a watchpoint

        :param      variable: The name of a variable or an address to watch
        :param bool write:    Write watchpoint
        :param bool read:     Read watchpoint
        """
        return self.protocols.execution.set_watchpoint(variable,
                                                  write=write,
                                                  read=read)

    @watch('TargetRemovebreakpoint')
    @action_valid_decorator_factory(TargetStates.STOPPED, 'execution')
    def remove_breakpoint(self, bkptno):
        """Deletes a breakpoint"""
        return self.protocols.execution.remove_breakpoint(bkptno)

    def update_state(self, state):
        self.log.info("State changed to to %s" % TargetStates(state))
        self.state = state

    @watch('TargetWait')
    def wait(self, state=TargetStates.STOPPED|TargetStates.EXITED):

        if state & TargetStates.SYNCING:
            self.log.warn("Waiting on SYNCING-state - this could lead to races")

        while True:
            if self.state & state != 0 and \
               (state & TargetStates.SYNCING or \
                self.state & TargetStates.SYNCING == 0):
                break
            time.sleep(.001) # send thread a ms to sleep to free resources

    def get_status(self):
        """
        Returns useful information about the target as a dict.
        """
        self.status['state'] = self.state
        return self.status

    def _get_unique_name(self, i=0):
        classname = type(self).__name__
        targetname = "{}{}".format(classname, i)
        if self.avatar and self.avatar.targets and targetname in self.avatar.targets:
            return self._get_unique_name(i + 1)
        return targetname

    def _resolve_executable_name(self):
        """
        Resolves the name of the executable for the endpoint.
        Order of operation:
            1: Check if config exists and whether target is installed
            2: Check sys_name from default config
            3: Check apt_name from default config
            4: BailOut
        """
        pass

    # ##generic aliases##
    wr = write_register
    rr = read_register
    rm = read_memory
    wm = write_memory
    bp = set_breakpoint



from types import MethodType

from capstone import *


def disassemble(self, addr=None, insns=1,
                arch=None, mode=None, detail=False):
    """
    Main purpose of the disassembler plugin, it's used to disassemble
    instructions
    :param addr:   The address to start disassembling.
                   If not specified, the current pc is used
    :param insns:  The numbe of instructions to be disassembled
    :param arch:   The capstone-architecture to be used.
                   If not specified, it is retrieved from avatar.arch
    :param mode:   The capstone-mode to be used.
                   If not specified, it is retrieved from avatar.arch
    :param detail: Whether to enable detailed disassembly
    :returns:     A list with capstone instructions
    """

    arch = self._arch.capstone_arch if not arch else arch
    mode = self._arch.capstone_mode if not mode else mode
    addr = self.regs.pc if not addr else addr

    ret = []
    md = Cs(arch, mode)
    md.detail = detail

    disassembled = 0
    while disassembled < insns:
        code_len = 8 * insns
        code = self.read_memory(addr, 1, code_len, raw=True)

        for ins in md.disasm(code, addr):
            ret.append(ins)
            disassembled += 1
            if disassembled >= insns:
                break

        addr += code_len
    return ret


def disassemble_pretty(self, addr=None, insns=1,
                       arch=None, mode=None):
    """
    Wrapper around disassemble to return disassembled instructions as string.
    """

    ret = ""
    disas = self.disassemble(addr, insns, arch, mode)

    for i in disas:
        ret += "0x%x:\t%s\t%s\n" % (i.address, i.mnemonic, i.op_str)

    return ret


def target_added_callback(avatar, *args, **kwargs):
    target = kwargs['watched_return']
    target.disassemble = MethodType(disassemble, target)
    target.disassemble_pretty = MethodType(disassemble_pretty, target)


def load_plugin(avatar):
    avatar.watchmen.add_watchman('AddTarget', when='after',
                                 callback=target_added_callback)
    for target in avatar.targets.values():
        target.disassemble = MethodType(disassemble, target)
        target.disassemble_pretty = MethodType(disassemble_pretty, target)



import sys
import logging
import struct
import ctypes

if sys.version_info < (3, 0):
    import Queue as queue
else:
    import queue

from math import ceil
from collections import OrderedDict
from os.path import abspath

# Python USB package used to interact with Inception-debugger
# Could be replace by the Inception driver
import usb.core
import usb.util

from avatar2.targets import TargetStates
from avatar2.message import AvatarMessage, UpdateStateMessage, BreakpointHitMessage


class InceptionProtocol(object):
    """
    This class implements the Inception protocol.
    It enables communication with the Inception-debugger hardware.

    :param additional_args:    Additional arguments delivered to Inception.
    :type  additional_args:    list
    :param device_vendor_id:   The usb device vendor id to connect to.
    :param device_product_id:  The usb device product id to connect to.
    """

    def __init__(self, additional_args=[], 
                 avatar=None, origin=None,
                 device_vendor_id=0x04b4, device_product_id=0x00f1,
                 output_directory='/tmp'):

        # USB device information
        self._device = None
        self._device_vendor_id = device_vendor_id
        self._device_product_id = device_product_id

        # pyusb device handler
        self._ep_out = None
        self._ep_in_response = None
        self._ep_in_irq = None

        # internal variables
        self._bkpt_limit = 0
        self._bkpt_list = [None] * 1
        self._debug_enabled = False

        self._queue = queue.Queue() if avatar is None \
                else avatar.queue
        self._fast_queue = queue.Queue() if avatar is None \
                else avatar.fast_queue

        self._origin = origin
        self.log = logging.getLogger('%s.%s' %
                                     (origin.log.name, self.__class__.__name__)
                                     ) if origin else \
            logging.getLogger(self.__class__.__name__)

    def shutdown(self):
        """
        Shuts down Inception

        returns: True on success
        """

        usb.util.dispose_resources(self._device)
        return True

    def connect(self):
        """
        Connects to USB3 Inception-debugger for all subsequent communication

        returns: True on success
        """

        self._device = usb.core.find(idVendor=self._device_vendor_id,
                idProduct=self._device_product_id)

        if self._device is None:
            self.log.critical('Failed to connect to Inception-debugger')
            raise ConnectionRefusedError("Inception-debugger is not connected")

        try:
            self._device.set_configuration()
            self._device.reset()
        except usb.core.USBError as e:
            self.log.critical("Could not set configuration: %s" % str(e))
            raise ConnectionRefusedError("Could not set configuration: %s" % str(e))

        # get an endpoint instance
        intf = self._device[0][(0,0)]

        if self._device.is_kernel_driver_active(0):
          self._device.detach_kernel_driver(0)

        self._ep_out = usb.util.find_descriptor(intf, bEndpointAddress=0x01)
        if self._ep_out is None:
            self.log.critical("Inception-debugger is connected but no endpoint 0x01 found")
            raise ConnectionRefusedError("Inception-debugger is connected but no endpoint 0x01 found")

        self._ep_in_response = usb.util.find_descriptor(intf, bEndpointAddress=0x81)
        if self._ep_in_response is None:
            self.log.critical("Inception-debugger is connected but no endpoint 0x81 found")
            raise ConnectionRefusedError("Inception-debugger is connected but no endpoint 0x81 found")
        
        self._ep_in_irq = usb.util.find_descriptor(intf, bEndpointAddress=0x82)
        if self._ep_in_irq is None:
            self.log.critical("Inception-debugger is connected but no endpoint 0x82 found")
            raise ConnectionRefusedError("Inception-debugger is connected but no endpoint 0x82 found")

        self.log.info("Connected to Inception-debugger")

        UpdateStateMessage(self, TargetStates.INITIALIZED)

        return True

    #def is_breakpoint_hitten(self):
    #    """
    #    Check if the processor hit a breakpoint

    #    :return:  PC or None if no breakpoint is hitten
    #    """

    #    self.stop()
    #    pc = self.read_pc()
    #    for i, j in enumerate(self._bkpt_list):
    #        if j == pc:
    #            self.log.info("Breakpoint hit")
    #            avatar_msg = BreakpointHitMessage(self._origin, bktp_nb, pc)
    #            self._queue.put(avatar_msg)
    #            return i, j
    #    self.cont()

    #    return None

    def reset(self):
        """
        Resets the target

        returns: True on success
        """
        pass

    def cont(self):
        """
        Resume the execution of the target

        :returns: True on success
        """
        pass

    def stop(self):
        """
        Stops the execution of the target

        :returns: True on success
        """
        pass

    def step(self):
        """
        Steps one instruction

        :returns:  True on success
        """
        pass

    def wait(self, state):
        """
        Wait until CPU reach provided state.

        :state:  state of the CPU
        """
        pass

    def check_halt(self):
        """
        Return if the CPU is halted or running.

        :return: True for halted, False for running
        """
        pass

    def read_pc(self):
        """
        Read PC register value
        The target must be stopped before reading the pc.

        :return:  program counter
        """
        pass

    def write_register(self, register, value):
        """
        Writing a register to the target

        :param register:     The name of the register
        :param value:        The actual value written to the register
        :return:             True on success
        """
        pass

    def read_register(self, register):
        """
        Reading a register from the target

        :param register:     The name of the register
        :return:             The actual value read from the register
        """
        pass

    def read_memory(self, address, wordsize=4, num_words=1, raw=False):
        """
        Read from memory of the target

        :param address:     The address to read from
        :param wordsize:    The size of a read word (1, 2, 4 or 8)
        :param words:       The amount of words to read (default: 1)
        :param raw:         Whether the read memory is returned unprocessed
        :return:            The read memory
        """
        pass

    def write_memory(self, address, wordsize, value, num_words=1, raw=False):
        """
        Writing to memory of the target

        :param address:   The address to write to
        :param wordsize:  The size of the write (1, 2, 4 or 8)
        :param value:     The actual value written to memory
        :type value:      int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """
        pass

    def set_breakpoint(self, address,
                       hardware=True, 
                       temporary=False, 
                       regex=False,
                       condition=None,
                       ignore_count=0,
                       thread=0,
                       pending=False):
        """Inserts a breakpoint

        :param bool hardware: Hardware breakpoint
        :param bool tempory:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with the condition
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        :return:              True on success
        """
        pass

    def remove_breakpoint(self, bkptnb):
        """
        Deletes a breakpoint

        :bkptnb:    Breakpoint number
        :return:    True on success
        """
        pass

    def set_watchpoint(self, variable, write=True, read=False):
        """Inserts a watchpoint

        :param variable:      The name of a variable or an address to watch
        :param bool write:    Write watchpoint
        :param bool read:     Read watchpoint
        :return:              True on success
        """
        pass


class IPCortexM3(InceptionProtocol):
    """
    Inception protocol for cortex-m3.
    """

    '''
    ARM documentation usefull for the Cortex M3 System Debug:
    * FPB: http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.ddi0337e/ch11s04s01.html
    * Core debug registers: http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.ddi0337e/ch11s04s01.html
    '''


    def __init__(self, additional_args=[], 
                 avatar=None, origin=None,
                 device_vendor_id=0x04b4, device_product_id=0x00f1,
                 output_directory='/tmp'):

        super(IPCortexM3, self).__init__(additional_args, 
                 avatar, origin, device_vendor_id, device_product_id, 
                 output_directory)

        self.regs = OrderedDict()
        self.regs.update({ "R0" :  0 })
        self.regs.update({ "R1" :  1 })
        self.regs.update({ "R2" :  2 })
        self.regs.update({ "R3" :  3 })
        self.regs.update({ "R4" :  4 })
        self.regs.update({ "R5" :  5 })
        self.regs.update({ "R6" :  6 })
        self.regs.update({ "R7" :  7 })
        self.regs.update({ "R8" :  8 })
        self.regs.update({ "R9" :  9 })
        self.regs.update({ "R10" : 10 })
        self.regs.update({ "R11" : 11 })
        self.regs.update({ "R12" : 12 })
        self.regs.update({ "SP" : 13 })
        self.regs.update({ "LR" : 14 })
        self.regs.update({ "PC" : 15 })
        self.regs.update({ "CPSR" : 16})

    def reset(self):
        """
        Resets the target

        returns: True on success
        """

        self.log.debug("Resetting target")

        # Reset JTAG
        data = '3000000030000000'
        self._ep_out.write(bytearray.fromhex(data))

        # Enable the FlashPatch module : breakpoint
        # FlashPatch Control Register (FP_CTRL)
        self.write_memory(0xE0002000, 4, 3)

        # Now we need to retrive the number of supported hw bkpt from the core
        FP_CTRL = self.read_memory(0xE0002000)

        # bits [7:4] are number of code slots field
        self._bkpt_limit = (FP_CTRL >> 4) & 0xF 
        self.log.debug(("Number of available breakpoints read %d") % (self._bkpt_limit))
        if self._bkpt_limit == 0:
            raise Exception("No hardware breakpoint found")

        # Watchpoint are located @ bits [11:8]
        #w = (FP_CTRL >> 8) & 0xF

        # Set SYSRESETREG bit at 1 in AIRCR register to request a system reset.
        # (system reset of all major components except for debug)
        self.write_memory(0xE000ED0C, 4, ((0xFA05 << 16) | (1 << 2)))

        return True

    def cont(self):
        """
        Resume the execution of the target

        :returns: True on success
        """

        self.log.debug("Attempted to continue execution on the target.")

        # Set C_HALT bit at 0 in the DHCSR register.
        self.write_memory(0xE000EDF0, 4, (0xA05F << 16) | 0b01)

        avatar_msg = UpdateStateMessage(self._origin, TargetStates.RUNNING)
        self._fast_queue.put(avatar_msg)

        return True

    def stop(self):
        """
        Stops the execution of the target

        :returns: True on success
        """

        self.log.debug("Attempted to stop execution of the target.")
        #DHCSR = self.read_memory(0xE000EDF0, 4)

        # Set C_HALT and C_DEBUGEN bits at 1 in DHCSR register
        self.write_memory(0xE000EDF0, 4, (0xA05F << 16) | 0b11)
        self._debug_enabled = True

        # Check the core is halted 
        DHCSR = self.read_memory(0xE000EDF0, 4)
        if not ((DHCSR >> 1) & 1):
            self.log.warning("Core not halted after stop")
        # Check the core acknowledges by reading S_HALT bit in DHCSR
        if not ((DHCSR >> 17) & 1):
            self.log.warning("Core not in debug state after stop")
            self._debug_enabled = False

        avatar_msg = UpdateStateMessage(self._origin, TargetStates.STOPPED)
        self._fast_queue.put(avatar_msg)

        return True

    def step(self):
        """
        Steps one instruction

        :returns:  True on success
        """

        self.log.debug("Attempted to step on the target.")

        # Enable Debug mode if not activated
        if not self._debug_enabled:
            self.write_memory(0xE000EDF0, 4, ((0xA05F << 16) | 0b11))
            self._debug_enabled = True

        # Check the core acknowledges by reading S_HALT bit in DHCSR
        DHCSR = self.read_memory(0xE000EDF0, 4)
        if not ((DHCSR >> 17) & 1):
            self.log.warning("Core is not in debug state before stepping")

        # Execute a step by setting the C_STEP bit to 1 in DHCSR register
        self.write_memory(0xE000EDF0, 4, (0xA05F << 16) | 0b101)

        avatar_msg = UpdateStateMessage(self._origin, TargetStates.RUNNING)
        self._fast_queue.put(avatar_msg)

        # Check the core is halted 
        DHCSR = self.read_memory(0xE000EDF0, 4)
        if not ((DHCSR >> 1) & 1):
            self.log.warning("Core is not halted after single step")

        avatar_msg = UpdateStateMessage(self._origin, TargetStates.STOPPED)
        self._fast_queue.put(avatar_msg)

        return True

    def wait(self, state):
        """
        Wait until CPU reach provided state.

        :state:  state of the CPU
        """

        halt = True if state == TargetStates.STOPPED else False
        
        self.log.debug("Wait until target reach %s state" % state)
        while (self.check_halt() != halt):
            pass

        avatar_msg = UpdateStateMessage(self._origin, state)
        self._fast_queue.put(avatar_msg)

    def read_register(self, register):
        """
        Reading a register from the target

        :param register:     The name of the register
        :return:             The actual value read from the register
        """

        self.log.debug("Read register %s" % register.upper())
        # Before reading the processor register value in DCRDR,
        # we must first ask to transfer it via DCRSR
        self.write_memory(0xE000EDF4, 4, self.regs[register.upper()])

        return self.read_memory(0xE000EDF8)

    def write_register(self, register, value):
        """
        Writing a register to the target

        :param register:     The name of the register
        :param value:        The actual value written to the register
        :return:    True on success
        """

        self.log.debug("Write register %s" % register.upper())
        # Write register value in DCRDR
        self.write_memory(0xE000EDF8, 4, value)
        # Ask DCRSR to transfer it to the processor register
        self.write_memory(0xE000EDF4, 4, (self.regs[register.upper()] |  (1 << 16)) )

        return True

    def read_memory(self, address, wordsize=4, num_words=1, raw=False):
        """
        Read from memory of the target

        :param address:     The address to read from
        :param wordsize:    The size of a read word (1, 2, 4 or 8)
        :param num_words:   The amount of words to read (default: 1)
        :param raw:         Whether the read memory is returned unprocessed
        :return:            The read memory
        """

        # Note: Inception does not support burst of read more than 4 bytes 

        # Command to order a read
        command = 0x24000001

        burst_size = 1
        start_addr = address

        # number of packets to be sent (header + payload)
        # 1 packet = 64 bits
        nb_packet = ceil((wordsize * num_words) / 4)
        nb_packet_init = nb_packet
        buf = ctypes.create_string_buffer(8 * burst_size)
        if raw:
            raw_ret = ctypes.create_string_buffer(nb_packet * 4)
        else:
            ret = []

        self.log.debug("Read_memory %s bytes starting from @=%s" % (nb_packet * 4, hex(address)))

        while (nb_packet > 0):

            burst = nb_packet if nb_packet < burst_size else burst_size
            nb_packet -= burst_size

            self.log.debug("Sending read for %s bytes from address @=%s" % (burst_size * 4, hex(start_addr)))
            # Create a burst
            for i in range(burst):
                struct.pack_into(">I", buf, (i * 8), command)
                struct.pack_into(">I", buf, (i * 8) + 4, start_addr)
                start_addr += 4

            self._ep_out.write(buf)

            # Unpack the bitstream of USB packets received
            for i in range(burst):

                response = self._ep_in_response.read(8, 0)
                # message is a bitstream of 64bits integer.
                # The highest 32bits are the status code
                if response[3] != 2:
                    raise Exception("Debugger returned an error")
                    
                value = struct.unpack_from(">I", response, 4)[0]
                self.log.debug("Read value: %s" % hex(value))

                if raw:
                    offset = (ceil((wordsize * num_words) / 4) - nb_packet - burst) * 4
                    struct.pack_into(">I", raw_ret, offset, value)
                else:
                    ret.append(value)

        if raw:
            return raw_ret.raw
        elif num_words == 1:
            return ret[0]
        else:
            return ret


    def write_memory(self, address, wordsize, value, num_words=1, raw=False):
        """
        Writing to memory of the target

        :param address:   The address to write to
        :param wordsize:  The size of the write (1, 2, 4 or 8)
        :param value:     The actual value written to memory
        :type value:      int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns:         True on success else False
        """

        # Command to order a write
        command = 0x14000001

        # Convert value to a list and compute the remainder bytes
        value, remainder = self._convert_to_list(wordsize, value, num_words, raw)

        # one packet is 12 bytes: 4 bytes commands, 
        # 4 bytes addresses and 4 bytes values
        nb_packet = len(value)
        buf = ctypes.create_string_buffer(12)
        struct.pack_into(">I", buf, 0, command)

        self.log.debug("Write_memory %s bytes starting from @=%s" % (nb_packet * 4, hex(address)))

        nb_full_packet = len(value) - 1 if remainder else nb_packet

        for i in range(0, nb_full_packet):
            val = value.pop(0)
            struct.pack_into(">I", buf, 4, address + i * 4)
            struct.pack_into(">I", buf, 8, val)

            self.log.debug("Write %s at address @=%s" % (hex(val), hex(address + i * 4)))
            self._ep_out.write(buf)

        # Write the remainder bytes when value is not multiple of 4 bytes
        if remainder:
            val = value.pop(0)
            self._write_remainder(remainder, address, val)

        return True

    def _convert_to_list(self, wordsize, value, num_words, raw=False):
        """
        Convert the value to be written into a list of 4 bytes and computes the remainder.

        :param wordsize:  The size of the write (1, 2, 4 or 8)
        :param value:     The actual value written to memory
        :type value:      int if num_words == 1 and raw == False
                          list if num_words > 1 and raw == False
                          str or byte if raw == True
        :param num_words: The amount of words to read
        :param raw:       Specifies whether to write in raw or word mode
        :returns: The value in a list of integer and the remainder 
        """

        remainder = 0

        if raw:
            if not len(value):
                raise ValueError("value has zero length")
            data = []

            for i in range(0, len(value), 4):         
                data.append(struct.unpack_from(">I", value, i)[0])

            remainder = len(value) % 4            
            if remainder == 1:
                data.append(struct.unpack_from(">B", value, len(value))[0])
            if remainder == 2:
                data.append(struct.unpack_from(">H", value, len(value))[0])

        elif num_words == 1:
            data = []
            data.append(value)
            remainder = wordsize % 4

        else:
            data = value[:]
            remainder = (wordsize * num_words) % 4

        if remainder == 3:
            raise ValueError("value must be 1 or multiple of 2")

        return data, remainder

    def _write_remainder(self, remainder, address, value):
        """
        When the value to be written is not a multiple of 4 bytes, 
        we must write the remainder by reading first the previous value.

        :param remainder: the number of remainder bytes to write
        :param address: the address where to write
        :param value: the value to write
        :type value: int
        :returns: True on success
        """

        command = 0x14000001
        buf = ctypes.create_string_buffer(12)

        if remainder == 3:
            raise ValueError("value must be 1 or multiple of 2")

        read_value = self.read_memory(address, 4, 1, raw=True)

        struct.pack_into(">I", buf, 0, command)
        struct.pack_into(">I", buf, 4, address)

        if remainder == 1:
            read_value = struct.unpack_from(">BBB", read_value, 0)
            #rem_value = struct.unpack_from(">B", struct.pack(">I", value), 0)[0]
            rem_value = struct.unpack_from(">B", value.raw, 0)[0]

            struct.pack_into(">B", buf, 8, read_value[0])
            struct.pack_into(">B", buf, 9, read_value[1])
            struct.pack_into(">B", buf, 10, read_value[2])
            struct.pack_into(">B", buf, 11, rem_value)
         
        if remainder == 2:
            read_value = struct.unpack_from(">H", read_value, 0)
            #rem_value = struct.unpack_from(">H", struct.pack(">I", value), 0)[0]
            rem_value = struct.unpack_from(">H", value, 0)[0]

            struct.pack_into(">H", buf, 8, read_value[0])
            struct.pack_into(">H", buf, 10, rem_value)

        self._ep_out.write(buf)

        return True

    def set_breakpoint(self, address,
                       hardware=False, 
                       temporary=False, 
                       regex=False,
                       condition=None,
                       ignore_count=0,
                       thread=0,
                       pending=False):
        """Inserts a breakpoint

        :param bool hardware: Hardware breakpoint
        :param bool tempory:  Tempory breakpoint
        :param str regex:     If set, inserts breakpoints matching the regex
        :param str condition: If set, inserts a breakpoint with the condition
        :param int ignore_count: Amount of times the bp should be ignored
        :param int thread:    Threadno in which this breakpoints should be added
        :return:              True on success
        """

        #if hardware == False:
        #    raise Exception("Software breakpoint not implemented")

        # Update bkpt counter and update bkpt register address
        indexes = [i for i, j in enumerate(self._bkpt_list) if j == None]

        # If no bkpt are available, raise an exception
        if indexes == []:
            raise Exception("Breakpoint limitation reaches")

        # FP_CTRL bit must also be set to enable breakpoints
        self.write_memory(0xE0002000, 4, 3)

        # Compute a free comparator register address
        FPCRegAddress = 0xE0002008 + ( indexes[0] * 4 )

        #set the flash patch comparator register value (FP_COMPx)
        FPCRegValue = 0
        FPCRegValue += 0b11 << 30 # Breakpoint on match on lower halfword
        FPCRegValue += address << 2 # Address to compare against
        FPCRegValue += 0b11 # Enable the comparator

        self.write_memory(FPCRegAddress, 4, FPCRegValue)
        self._bkpt_list[indexes[0]] = address
        self.log.info("Breakpoint set")

        return True

    def set_watchpoint(self, variable, write=True, read=False):
        """Inserts a watchpoint
        Not implemented yet

        :param      variable: The name of a variable or an address to watch
        :param bool write:    Write watchpoint
        :param bool read:     Read watchpoint
        :return:    True on success
        """

        self.log.critical("Watchpoint not implemented")
        raise Exception("Watchpoint not implemented")
        return False

    def remove_breakpoint(self, bkptnb):
        """
        Deletes a breakpoint

        :bkptnb:    Breakpoint number
        :return:    True on success
        """

        if self._bkpt_limit < bkptnb :
            raise Execption("bkptnb higher than supported breakpoint")

        # Update bkpt counter and update bkpt register address
        FPCRegAddress = 0xE0002008 + ( bkptnb * 4 )

        #set the flash patch comparator register value (FP_COMPx)
        FPCRegValue += b'0x00' # Enable the comparator

        write_memory(self, FPCRegAddress, 4, FPCRegValue)

        self._bkpt_list[bkptnb] = None
        self.log.info("Breakpoint removed")

        return True

    def check_halt(self):
        """
        Return if the CPU is halted or running.

        :return: True for halted, False for running
        """

        halted = False

        # Check the C_HALT bit in the DHCSR register.
        DHCSR = self.read_memory(0xE000EDF0)
        if ((DHCSR >> 1) & 1):
              halted = True

        return halted

    def read_pc(self):
        """
        Read PC register value.
        The target must be stopped before reading the pc.

        :return:  pc
        """

        # Read pc
        DCRDR = self.read_register("PC")
        #PSR = self.read_register("CPSR")

        # We are always in Thumb mode with Cortex M3
        #return DCRDR - 4
        return DCRDR 




from setuptools import setup
from sys import version_info


setup(
    name='avatar2',
    version='1.4.5',
    packages=['avatar2',
              'avatar2/archs',
              'avatar2/targets',
              'avatar2/protocols',
              'avatar2/peripherals',
              'avatar2/plugins',
              'avatar2/plugins/arm',
              'avatar2/installer'
              ],
    install_requires=[
        'enum34;python_version<"3.4"',
        'pygdbmi==0.9.0.2;python_version<="3.5"',
        'pygdbmi>=0.10.0.0;python_version>"3.5"',
        'cached_property;python_version<="3.8"',
        'intervaltree',
        'posix_ipc>=1.0.0',
        'capstone>=3.0.4',
        'keystone-engine',
        'parse',
        'configparser',
        'npyscreen',
        'unicorn',
        'bitstring',
        'pylink-square',
        'pyusb',
    ],
    include_package_data=True,
    url='https://github.com/avatartwo/avatar2',
    description='A Dynamic Multi-Target Orchestration Framework',
    maintainer='Marius Muench',
    maintainer_email='marius.muench@eurecom.fr'
)



from types import MethodType

#from keystone import *
from ...targets.unicorn_target import UnicornTarget
from intervaltree import IntervalTree

from unicorn import UC_HOOK_MEM_READ, UC_HOOK_MEM_WRITE, UC_MEM_READ, UC_MEM_WRITE

def _callable_uc_mem_hook(self, uc, access, address, size, value, user_data):

    # Sync on access. However, we only syncronize pages which are not forwarded
    if self.avatar.get_memory_range(address).forwarded is False and \
       self._callable_synced_pages[address] == set():
        page = address - (address % 0x1000)
        content = self.read_memory( page, 0x1000, raw=True)
        uc.mem_write(page, content)
        self._callable_synced_pages[page:page+0x1000] = content

    if access == UC_MEM_WRITE:
        self._callable_writes.append((address, size, value)) 
        if self._callable_do_writes is True:
            self.write_memory(address, size, value)




def call(self, address, args=None, playground=0x10000000, do_writes=False):
    '''
    Calls the function at the given address with the current state of the program.
    Internally, this method creates a unicorn target

    :returns: a dictionary with changed memory locations and values.
    '''

    # Step1: Set up unicorn
    self._callable_writes = []
    self._callable_do_writes = True
    self._callable_synced_pages = IntervalTree()
    
    if hasattr(self, '_callable_unicorn'):
        unicorn = self._callable_unicorn
    else:
        unicorn = UnicornTarget(self.avatar)
        self._callable_unicorn = unicorn

    unicorn.init()
    unicorn.wait()

    unicorn.add_hook(UC_HOOK_MEM_READ | UC_HOOK_MEM_WRITE,
                     self._callable_uc_mem_hook)
    unicorn.protocols.memory.uc.mem_map(playground, 0x1000)
    

    # Step2: Setup Unicorn State
    self.avatar.transfer_state(self, unicorn)
    unicorn.set_breakpoint(unicorn.regs.lr)

    # The following sets up the args in memory
    playground_pointer = playground
    for i in range(len(args)):
        if type(args[i]) == int:
            setattr(unicorn.regs, 'r%d' % i, args[i]) 
        unicorn.write_memory(playground_pointer, 1, args[i], raw=True)
        playground_pointer += len(args[i])


    # Step3: Do the emulation
    self.log.info("Starting Emulation of %x" % address)
    unicorn.regs.pc = address
    unicorn.cont()
    unicorn.wait()
    

    # Step4: Cleanup & Exit
    writes = self._callable_writes
    del(self._callable_writes)
    del(self._callable_do_writes)
    del(self._callable_synced_pages)

    return writes
    
    




def target_added_callback(avatar, *args, **kwargs):
    target = kwargs['watched_return']
    target.call = MethodType(call, target)
    target._callable_uc_mem_hook = MethodType(_callable_uc_mem_hook, target)


def load_plugin(avatar):
    avatar.watchmen.add_watchman('AddTarget', when='after',
                                 callback=target_added_callback)
    for target in avatar.targets.values():
        target.call = MethodType(call, target)
        target._callable_uc_mem_hook = MethodType(_callable_uc_mem_hook, target)



from __future__ import print_function
from distutils.spawn import find_executable as find
from os import system, chdir
from os.path import exists
from time import sleep
from shutil import rmtree
from curses import endwin as get_terminal_screen
from .config import *

import npyscreen as nps




# Strings used in varius menues
WELCOME_DIALOG = ('Welcome to the avatar2 target install system.\n\n'
                  'This installer will store settings into %s.\n'
                  'Please select \'ok\' if you agree.' % CONFIG_FILE)

ALLDONE_DIALOG = ('All requested targets are installed.\n'
                  'Returning to main menu')

ALR_INSTALL_WARN = ('It seems %s is already installed at %s\n'
                    'Are you sure you want to continue?') 

DIR_EXISTS_WARN = ('Installation directory %s exists.\n'
                   'Do you want to delete it?')

INSTALL_FAILED_ERR = ('Installation of %s failed!\n'
                      'Error: %s \n')

INSTALL_SUCCESS = ('Installation of %s succeeded!\n')


VERIFY_GIT_INSTALL = ('About to install target from git.\n'
                      'Please make sure the settings below are correct.')



MENTRY_GIT_INSTALL = 'Install via git'
MENTRY_APT_INSTALL = 'Install via apt-get'
MENTRY_CHANGE_PATH = 'Set path for installed binary'
MENTRY_CANCEL = 'Cancel'
MENTRY_BUILD = 'Build Target!'
MENTRY_FETCH_DEPS = 'Install dependencies via apt-get'






class AvatarInstallerMenu(nps.Form):
    EXTRA_KWARGS = []

    def __init__(self, *args, **keywords):
        self.next_form = keywords.get('next_form')
        self.action_on_ok = keywords.get('action_on_ok')

        [setattr(self, key, keywords.get(key)) for key in self.__class__.EXTRA_KWARGS]

        super(nps.Form, self).__init__(*args, **keywords)
        if self.name is None:
            self.name = 'avatar2 target installer'


class AvatarInstallerTargetMenu(AvatarInstallerMenu):

    def create(self):
        self.target_name = self.parentApp.current_target
        self.name = self.target_name + ' installer'
        options = []
        config = self.parentApp.config

        if TARGETS[self.target_name].get('git') is not None:
            options.append( MENTRY_GIT_INSTALL )

        if TARGETS[self.target_name].get('apt_name') is not None and \
           config.getboolean('DIST', 'has_apt') is True:
            options.append( MENTRY_APT_INSTALL )

        options.append( MENTRY_CHANGE_PATH )
        options.append( MENTRY_CANCEL)

        self.opt_form = self.add(nps.TitleSelectOne,
                                 name='What do you want to do?',
                                 values = options, value=[0,],
                                 scroll_on_exit=True
                                )

    def afterEditing(self):
        sel = self.opt_form.get_selected_objects()
        if len(sel) == 0:
            return
        sel = sel[0]

        if sel == MENTRY_CANCEL:
            self.parentApp.install_next_target()
        elif sel == MENTRY_APT_INSTALL:
            apt_name = TARGETS[self.target_name]['apt_name']
            if self.parentApp.apt_install(apt_name) is True:
                self.parentApp.setNextForm('InstallerSuccess')
            else:
                self.parentApp.setNextForm('InstallationFailed')

        elif sel == MENTRY_GIT_INSTALL:
            self.parentApp.setNextForm('GitInstaller')
        elif sel == MENTRY_CHANGE_PATH:
            self.parentApp.setNextForm('PathChanger')

class AvatarInstallerGitMenu(AvatarInstallerMenu):
    def create(self):
        self.target_name = self.parentApp.current_target
        self.name = self.target_name + ' git installer'

    
        self.form = self.add(nps.TitleText, name=VERIFY_GIT_INSTALL,
                             autowrap=True, editable=False)

        self.install_dir = self.parentApp.config.get(
            'DIST', 'default_install_path') + self.target_name

        self.git_path = self.add(nps.TitleText, name='Remote Repository',
                                value=TARGETS[self.target_name]['git'],
                                begin_entry_at=25)

        self.git_branch = self.add(nps.TitleText, name='Remote Branch',
                                   value='master', begin_entry_at=25)

        self.install_dir = self.add(nps.TitleFilename,
                                    name='Local directory',
                                    value=self.install_dir,
                                    begin_entry_at=25)
        
        conf = TARGETS[self.target_name].get('configure')
        self.configure_options = self.add(
             nps.TitleText, name='Configure options', begin_entry_at=25,
             value=conf, hidden=False if conf is not None else True
        )

        script = TARGETS[self.target_name].get('install_script')
        self.install_script = self.add(
                nps.TitleText, name='Install script (if used following parameters are ignored)',
                value=script, begin_entry_at=25,
                hidden=False if script is not None else True
        )

        make = TARGETS[self.target_name].get('make')
        self.make_options= self.add(
            nps.TitleText, name='Make options', begin_entry_at=25,
            value=make, hidden=False if make is not None else True
        )

        self.git_apt_deps = self.add(
                nps.TitleText, name='Dependencies with apt-get (default values for Ubuntu 20.04)',
                value=TARGETS[self.target_name]['git_apt_deps'],
                begin_entry_at=25
        )
           

        options = [MENTRY_BUILD, MENTRY_CANCEL]
        self.opt_form = self.add(nps.TitleSelectOne,
                                 name='What do you want to do?',
                                 values=options, value =[0,],
                                 scroll_on_exit=True)

    def afterEditing(self):
        sel = self.opt_form.get_selected_objects()
        if len(sel) == 0:
            return
        sel = sel[0]

        if sel == MENTRY_CANCEL:
            self.parentApp.setNextFormPrevious()
        if sel == MENTRY_BUILD:
            self.parentApp.install_dir = self.install_dir.value
            
            # save all the values in case we need to delete a dir
            self.parentApp.git_args = (self.install_dir.value,
                                       self.git_path.value,
                                       TARGETS[self.target_name]['install_cmd'])
            self.parentApp.git_kwargs = {
                'branch': self.git_branch.value,
                'apt_deps': self.git_apt_deps.value,
                'configure_options': self.configure_options.value,
                'make_options': self.make_options.value,
                'install_script': self.install_script.value
            }
            if exists(self.install_dir.value):
                self.parentApp.switchForm('DirExists')
                return
            # In the long run, we should aim for a seperated 'RunInstall'-form
            try:
                self.parentApp.git_install(*self.parentApp.git_args,
                                           **self.parentApp.git_kwargs)
                self.parentApp.setNextForm('InstallerSuccess')
            except Exception as e:
                self.parentApp.error = e
                self.parentApp.setNextForm('InstallationFailed')

class AvatarInstallerInstallError(nps.Popup,AvatarInstallerMenu):
    def create(self):
        self.text=INSTALL_FAILED_ERR % (self.parentApp.current_target,
                                        self.parentApp.error)
        self.add(nps.Pager, values=self.text.split('\n'),
                 autowrap=True, editable=False)

    def afterEditing(self):
        self.parentApp.setNextForm('TargetInstaller')


class AvatarInstallerInstallSuccess(nps.Popup,AvatarInstallerMenu):
    def create(self):
        self.text=INSTALL_SUCCESS % self.parentApp.current_target
        self.add(nps.Pager, values=self.text.split('\n'),
                 autowrap=True, editable=False)

    def afterEditing(self):
        self.parentApp.install_next_target()


class AvatarInstallerAF(nps.ActionFormV2, AvatarInstallerMenu):
    EXTRA_KWARGS = ['exit_on_cancel']

    def on_ok(self):
        if self.action_on_ok:
            res = self.action_on_ok()
        if self.next_form:
            self.parentApp.setNextForm(self.next_form)

    def on_cancel(self):
        if self.exit_on_cancel is True:
            exit()
        else:
            self.parentApp.setNextFormPrevious()


class AvatarInstallerUpdatePath(AvatarInstallerAF):

    def create(self):
        self.target_name = self.parentApp.current_target
        self.name = self.target_name + ' installer'
        self.install_dir = self.add(
            nps.TitleFilenameCombo,
            name='Choose new location:',
            value=self.parentApp.config.get('TARGETS', self.target_name)
        )

    def on_ok(self):
        self.parentApp.config.set('TARGETS', self.target_name,
                                  self.install_dir.value)
        self.parentApp.install_next_target()


class AvatarInstallerWarning(AvatarInstallerAF):
    EXTRA_KWARGS = AvatarInstallerAF.EXTRA_KWARGS + ['text']

    def create(self):
        self.add(nps.Pager, values=self.text.split('\n'),
                 autowrap=True, editable=False)
    
class AvatarInstallerWarningPopup(nps.ActionPopup, AvatarInstallerWarning):
    ''' Displays a warning as popup '''
    pass

class AvatarInstallerWarningAlreadyInstalled(AvatarInstallerWarningPopup):
    def create(self):
        path = self.parentApp.config.get_target_path(self.parentApp.current_target)
        self.name='WARNING!'
        self.text=ALR_INSTALL_WARN % (self.parentApp.current_target, path)
        super(AvatarInstallerWarningPopup, self).create()

class AvatarInstallerWarningDirExists(AvatarInstallerWarningPopup):
    def create(self):
        self.name='WARNING!'
        self.text= DIR_EXISTS_WARN % (self.parentApp.install_dir)
        super(AvatarInstallerWarningDirExists, self).create()


    def on_ok(self):
        self.parentApp.delete_dir()
        try:
            self.parentApp.git_install(*self.parentApp.git_args,
                                       **self.parentApp.git_kwargs)
            self.parentApp.setNextForm('InstallerSuccess')
        except Exception as e:
            self.parentApp.error = e
            self.parentApp.setNextForm('InstallationFailed')


class AvatarInstallerTargetSelector(AvatarInstallerAF):
    OK_BUTTON_BR_OFFSET = (2, 14)
    CANCEL_BUTTON_BR_OFFSET = (2, 6)
    OK_BUTTON_TEXT          = "OK"
    CANCEL_BUTTON_TEXT = "EXIT"

    def create(self):
        self.exit_on_cancel = True
        self.opt_form = self.add(
            nps.TitleMultiSelect,
            name='Which Targets do you want to install/modify?',
            values=list(TARGETS.keys()))

    def on_ok(self):
        self.parentApp.installer_list = self.opt_form.get_selected_objects()
        if self.parentApp.installer_list is not None:
            self.parentApp.current_target = self.parentApp.installer_list.pop()
            path = self.parentApp.config.get_target_path(self.parentApp.current_target)
            if path == None:
                self.parentApp.switchForm('TargetInstaller')
            else:
                self.parentApp.setNextForm('AlreadyInstalled')


class Avatar2Installer(nps.NPSAppManaged):
    '''
    The installer holds all the logic to install targets.
    It create the UI forms in its onStart method, all other functions
    are just callbacks called form the UI

    For dynamic content in windows, all neccessary information are saved
    in the AvatarInstaller object, which  also can be written by the forms.
    As a rule of thumb, forms should NOT save any data on their own.
    '''

    STARTING_FORM = 'WelcomeDialog'



    def install_next_target(self):
        self.config.write_config()
        if len( self.installer_list ) > 0:
            self.current_target = self.installer_list.pop()
            self.setNextForm('TargetInstaller')
        else:
            self.setNextForm('AllDone')

    def apt_install(self, package):
        get_terminal_screen()
        res = system('sudo apt-get install %s' % package)
        sleep(1.5)
        if res == 0:
            self.config.set('TARGETS', self.current_target, package)
            return True
        else:
            return False

    def git_install(self, local_directory, repository, install_commands,
                    branch='master', apt_deps=None, configure_options=None,
                    make_options=None, install_script=None):

        get_terminal_screen()

        if apt_deps != '':
            system('echo')
            system('echo [*] Installing dependencies...')
            res = system('sudo apt-get install %s' % apt_deps)
            sleep(1.5)
            if res != 0:
                raise Exception('Fail to install dependencies %s' % apt_deps)

        system('echo')
        system('echo [*] Installing from git...')
        git_exec = find('git')
        system('%s clone %s --single-branch --branch %s %s'  % 
               (git_exec, repository, branch, local_directory) )

        chdir(local_directory)

        if install_script is None or install_script.strip() == '':
            for cmd in install_commands:
                if cmd == './configure' and configure_options is not None:
                    res = system(cmd + ' ' + configure_options)
                elif cmd == 'make' and make_options is not None:
                    res = system(cmd + ' ' + make_options)
                else:
                    res = system(cmd)
                if res != 0:
                    sleep(2)
                    raise Exception('Executing install command \'%s\' failed' % cmd)

        else:
            res = system('echo ' + install_script + ' ' + configure_options)
            res = system(install_script + ' ' + configure_options)
            if res != 0:
                raise Exception('Install script %s failed' % install_script)

        exec_path = '%s/%s' % (self.install_dir,
                               TARGETS[self.current_target]['rel_path'])
        if exists(exec_path):
            self.config.set('TARGETS', self.current_target, exec_path)

    def delete_dir(self):
        rmtree(self.install_dir)
        
    def onStart(self):

        self.config = AvatarConfig()
        self.current_target = None
        self.install_dir = None
        self.addForm('WelcomeDialog', AvatarInstallerWarningPopup, 
                     text=WELCOME_DIALOG, exit_on_cancel=True,
                     next_form='TargetSelector',
                     action_on_ok=self.config.write_config)

        self.addForm('TargetSelector', AvatarInstallerTargetSelector,
                     next_form='AlreadyInstalled')

        # Forms with dynamically generated content require a formclass,
        # which will spawn a new FormInstance on every edit of the form
        self.addFormClass('AlreadyInstalled',
                          AvatarInstallerWarningAlreadyInstalled,
                          next_form='TargetInstaller')

        self.addFormClass('InstallationFailed', AvatarInstallerInstallError)
        self.addFormClass('TargetInstaller', AvatarInstallerTargetMenu)
        self.addFormClass('GitInstaller', AvatarInstallerGitMenu)
        self.addFormClass('DirExists', AvatarInstallerWarningDirExists,
                          action_on_exit=self.delete_dir)
        self.addFormClass('PathChanger', AvatarInstallerUpdatePath)
        self.addFormClass('InstallerSuccess', AvatarInstallerInstallSuccess)

        self.addForm('AllDone', AvatarInstallerWarningPopup, 
                     text=ALLDONE_DIALOG, exit_on_cancel=True,
                     next_form='TargetSelector')


if __name__ == '__main__':
   Avatar2Installer().run()



from intervaltree.intervaltree import IntervalTree
from types import MethodType
from threading import Event
from enum import Enum

from avatar2.watchmen import AFTER, BEFORE, watch
from avatar2 import TargetStates, GDBTarget


def load_memory_mappings_target(target, *args, **kwargs):
    """
    Stub method for compatibility to be able to call load_memory_mappings on the
    target object.
    """
    return load_memory_mappings(target.avatar, target, *args, **kwargs)


def load_memory_mappings(avatar, target, forward=False, update=True):
    """
    Load memory maps from the specified target
    :param forward: Enable forwarding of memory to that target
    :param update:  If true, replaces avatars memory_ranges with the loaded ones
    :return:        An Intervaltree object containing the mappings
    """
    if not isinstance(target, GDBTarget):
        raise TypeError("The memory mapping can be loaded ony from GDBTargets")

    ret, resp = target.protocols.execution.get_mappings()
    lines = resp.split("\n")[4:]
    mappings = [
        {
            "start": int(x[0], 16),
            "end": int(x[1], 16),
            "size": int(x[2], 16),
            "offset": int(x[3], 16),
            "obj": x[4],
        }
        for x in [y.split() for y in lines]
    ]
    memory_ranges = IntervalTree()

    for m in mappings:
        avatar.add_memory_range(
            m["start"],
            m["size"],
            name=m["obj"],
            forwarded=forward,
            forwarded_to=target if forward else None,
            interval_tree=memory_ranges,
        )
    if update is True:
        avatar.memory_ranges = memory_ranges
    return memory_ranges


def add_methods(target):
    target.load_memory_mappings = MethodType(load_memory_mappings_target, target)


def target_added_callback(avatar, *args, **kwargs):
    target = kwargs["watched_return"]
    add_methods(target)


def load_plugin(avatar):
    avatar.watchmen.add_watchman(
        "AddTarget", when="after", callback=target_added_callback
    )
    avatar.load_memory_mappings = MethodType(load_memory_mappings, avatar)
    for target in avatar.targets.values():
        add_methods(target)



from avatar2.targets import Target, TargetStates
from avatar2.protocols.inception import InceptionProtocol, IPCortexM3

from avatar2.watchmen import watch

class InceptionTarget(Target):
    '''
    Inception is a framework to perform security testing of real world
    firmware programs. The inception target and protocol allow Avatar2
    to interact with the low latency inception debugger and speed up
    read/write operations between the host and the device.

    For more information, please visit:
    https://inception-framework.github.io/inception/debugger.html

    Publication:
    Inception: System-wide security testing of real-world embedded systems
    software
    Nassim Corteggiani, Giovanni Camurati, Aurélien Francillon
    27th USENIX Security Symposium (USENIX Security 18), Baltimore, MD
    '''

    def __init__(self, avatar, 
                 processor='cortex-m3',
                 device_vendor_id=0x04b4,
                 device_product_id=0x00f1,
                 **kwargs):

        super(InceptionTarget, self).__init__(avatar, **kwargs)

        self.processor = processor
        self._device_product_id = device_product_id
        self._device_vendor_id = device_vendor_id

    @watch('TargetInit')
    def init(self):

        if self.processor == 'cortex-m3':
            inception = IPCortexM3(avatar=self.avatar, origin=self,
                    device_vendor_id=self._device_vendor_id,
                    device_product_id=self._device_product_id, 
                    output_directory=self.avatar.output_directory)
        else:
            inception = None
            self.log.warning("Target board not implemented")
            raise Exception("Target board not implemented")


        if inception.connect():
            inception.reset()
            self.update_state(TargetStates.RUNNING)
            self.log.info("Connected to Target")
        else:
            self.log.warning("Connecting failed")
            raise Exception("Connecting to target failed")

        if inception.stop():
            self.update_state(TargetStates.STOPPED)

        self.protocols.set_all(inception)
        self.protocols.monitor = inception 
        
        #self.wait()

    def reset(self):
        return self.protocols.execution.reset()

    @watch('TargetWait')
    def wait(self, state=TargetStates.STOPPED):
        return self.protocols.execution.wait(state)




#/usr/bin/env 

import logging
import re
import socket
import binascii
import struct

import xml.etree.ElementTree as ET

from time import sleep
from struct import pack
from types import MethodType
from threading import Thread, Event
from socket import AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR
from os.path import dirname

from avatar2.targets import TargetStates

l = logging.getLogger('avatar2.gdbplugin')

chksum = lambda x: sum(x) & 0xff
match_hex = lambda m, s: [int(x, 16) for x in re.match(m, s).groups()]

TIMEOUT_TIME = 1.0


class GDBRSPServer(Thread):

    def __init__(self, avatar, target, port=3333, xml_file=None,
                 do_forwarding=False):
        super().__init__()
        self.daemon=True
        self.sock = socket.socket(AF_INET, SOCK_STREAM)
        self.sock.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

        self.avatar = avatar
        self.target = target
        self.port = port
        self.xml_file = xml_file
        self.do_forwarding = do_forwarding

        self._packetsize=0x47FF
        self.running = False
        self.bps = {}
        self._do_shutdown = Event()
        

        xml_regs = ET.parse(self.xml_file).getroot().find('feature')
        self.registers = [reg.attrib for reg in xml_regs if reg.tag == 'reg']
        assert(len(self.registers))

        self.handlers = {
            'q' : self.query,
            'v' : self.multi_letter_cmd,
            'H' : self.set_thread_op,
            '?' : self.halt_reason,
            'g' : self.read_registers,
            'G' : self.reg_write,
            'm' : self.mem_read,
            'M' : self.mem_write,
            'c' : self.cont,
            'C' : self.cont, #cond with signal, we don't care
            's' : self.step,
            'S' : self.step,
            'S' : self.step_signal,
            'Z' : self.insert_breakpoint,
            'z' : self.remove_breakpoint,
            'D' : self.detach,
        }

    def shutdown(self):
        self._do_shutdown.set()
        sleep(TIMEOUT_TIME*2)

    def run(self):

        l.info(f'GDB server listening on port {self.port}, please connect')
        self.sock.bind(('', self.port))
        self.sock.settimeout(TIMEOUT_TIME)
        self.sock.listen(1)
        
        while not self._do_shutdown.isSet():
            try:
                self.conn, addr = self.sock.accept()
            except socket.timeout:
                continue
            self.conn.settimeout(TIMEOUT_TIME)
            l.info(f'Accepted connection from {addr}')

            if not self.target.state & TargetStates.STOPPED:
                self.target.stop()
            while self.conn._closed is False:
                packet = self.receive_packet()
                if packet is None:
                    continue

                l.debug(f'Received: {packet}')
                self.send_raw(b'+') # send ACK

                handler = self.handlers.get(chr(packet[0]),
                                                self.not_implemented)
                resp = handler(packet)
                if resp is not None:
                    self.send_packet(resp)
        self.sock.close()


    ### Handlers
    def not_implemented(self, pkt):
        l.critical(f'Received not implemented packet: {pkt}')
        return b''

    def query(self, pkt):
        if pkt[1:].startswith(b'Supported') is True:
            feat = [b'PacketSize=%x' % self._packetsize,
                    b'qXfer:features:read+'
                   ]
            return b';'.join(feat)

        if pkt[1:].startswith(b'Attached') is True:
            return b'1'

        if pkt[1:].startswith(b'Xfer:features:read:target.xml') is True:
            off, length = match_hex('qXfer:features:read:target.xml:(.*),(.*)',
                                   pkt.decode())
            
            with open(self.xml_file, 'rb') as f:
                data = f.read()
            resp_data = data[off:off+length]
            if len(resp_data) < length:
                prefix = b'l'
            else:
                prefix = b'm'
            return prefix+resp_data

        if pkt[1:].startswith(b'fThreadInfo') is True:
            return b'm1'
        if pkt[1:].startswith(b'sThreadInfo') is True:
            return b'l'

        if pkt[1:].startswith(b'Rcmd') is True: # Monitor commands
            try:
                cmd = re.match('qRcmd,(.*)',pkt.decode())[1]
                cmd = binascii.a2b_hex(cmd) 
                l.debug(f'Receiced cmd: {cmd}')
                res = eval(cmd)
                
                self.send_packet(b'O' \
                            + binascii.b2a_hex(repr(res).encode()) \
                            + b'0a')
                return b'OK'
                
            except Exception as e:
                self.send_packet(b'O' + b'ERROR: '.hex().encode())
                
                if hasattr(e, 'msg'):
                    self.send_packet(b'O' \
                                + e.msg.encode().hex().encode() \
                                + b'0a')
                elif hasattr(e, 'args'):
                    self.send_packet(b'O' \
                                + e.args[0].encode().hex().encode() \
                                + b'0a')
                    
                return b'OK'

        return b''

    def multi_letter_cmd(self, pkt):
        if pkt[1:].startswith(b'vMustReplyEmpty') is True:
            return b''
        return b''

    def set_thread_op(self, pkt):
        return b'OK' # we don't implement threads yet

    def halt_reason(self, pkt):
        return b'S00' # we don't specify the signal yet

    def read_registers(self, pkt):
        resp = ''
        for reg in self.registers:
            
            bitsize = int(reg['bitsize'])
            assert( bitsize % 8 == 0)
            r_len = int(bitsize / 8)
            r_val = self.target.read_register(reg['name'])
            #l.debug(f'{reg["name"]}, {r_val}, {r_len}')

            resp += r_val.to_bytes(r_len, 'little').hex()
            
        return resp.encode()
    
    def reg_write(self, pkt):
        idx = 1 # ignore the first char of pkt
        for reg in self.registers:
            bitsize = int(reg['bitsize'])
            r_len = int(bitsize / 8)
            r_val = pkt[idx: idx + r_len*2]
            r_raw = bytes.fromhex(r_val.decode())
            int_val =  int.from_bytes(r_raw, byteorder='little')

            self.target.write_register(reg['name'], int_val)
            idx += r_len*2
        return b'OK'


    def mem_read(self, pkt):
        try:
            addr, n = match_hex('m(.*),(.*)', pkt.decode())

            if self.do_forwarding is True:
                mr = self.avatar.get_memory_range(addr)
                if mr is not None and mr.forwarded is True:
                    val = mr.forwarded_to.read_memory(addr, n)
                    val = val.to_bytes(n, byteorder='little')
                    return binascii.b2a_hex(val)

            val = self.target.read_memory(addr, n, raw=True).hex()
            return val.encode()
            
        except Exception as e:
            l.warn(f'Error in mem_read: {e}')
            return b'E00'


    def mem_write(self, pkt):
        try:
            addr, n, val = match_hex('M(.*),(.*):(.*)', pkt.decode())
            raw_val = val.to_bytes(n, byteorder='big') # wtf :/

            if self.do_forwarding is True:
                mr = self.avatar.get_memory_range(addr)
                if mr is not None and mr.forwarded is True:
                    int_val = int.from_bytes(raw_val,byteorder='little')
                    mr.forwarded_to.write_memory(addr, n, int_val)
                    return b'OK'

            self.target.write_memory(addr, n, raw_val, raw=True)
            return b'OK'
            
        except Exception as e:
            l.warn(f'Error in mem_write: {e}')
            return b'E00'


    def cont(self, pkt):
        self.target.cont()
        self.running = True
        return b'OK'

    def step(self, pkt):
        self.target.step()
        return b'S00'

    def step_signal(self, pkt):
        self.target.step()
        return pkt[1:]

    def insert_breakpoint(self, pkt):
        addr, kind = match_hex('Z0,(.*),(.*)', pkt.decode())
        bpno = self.target.set_breakpoint(addr)
        self.bps[bpno] = addr
        return b'OK'

    def remove_breakpoint(self, pkt):
        addr, kind = match_hex('z0,(.*),(.*)', pkt.decode())
        matches = []
        for n, a in self.bps.items():
            if a == addr:
                matches.append(n)
        if len(matches) == 0:
            l.warn(f'GDB tried to remove non existing bp for {addr}')
            l.info(self.bps)
            return b'E00'
        
        self.target.remove_breakpoint(n)
        self.bps.pop(n)
        return b'OK'

    def detach(self, pkt):
        l.info("Exiting GDB server")
        if not self.target.state & TargetStates.EXITED:
            for bpno in self.bps.items():
                self.target.remove_breakpoint(bpno)
            self.target.cont()
        if self.conn._closed is False:
            self.send_packet(b'OK')
            self.conn.close()
        
        return None

    ### Sending and receiving

    def send_packet(self, pkt):
        if type(pkt) == str:
            raise Exception("Packet require bytes, not strings")
        
        self.send_raw(b'$%b#%02x' % (pkt, chksum(pkt)))


    def send_raw(self, raw_bytes):
        l.debug(f'Sending data: {raw_bytes}')
        self.conn.send(raw_bytes)


    def check_breakpoint_hit(self):
        if self.target.state & TargetStates.STOPPED and self.running is True:
            if self.target.regs.pc in self.bps.values():
                self.running = False
                self.send_packet(b'S05')


    def receive_packet(self):
        pkt_finished = False
        pkt_receiving = False
        while pkt_finished is False:
            try:
                c = self.conn.recv(1)
            except socket.timeout:
                if self._do_shutdown.isSet():
                    self.send_packet(b'S03')
                    self.conn.close()
                    return

                if self.target.state & TargetStates.EXITED:
                    self.send_packet(b'S03')
                    self.conn.close()
                    return
                self.check_breakpoint_hit()
                continue

            if c == b'\x03':
                if not self.target.state & TargetStates.STOPPED:
                    self.target.stop()
                self.send_packet(b'S02')
            elif c == b'$': # start of package
                pkt = b''
                pkt_receiving = True
            elif c == b'#': # end of package
                checksum = self.conn.recv(2)
                if int(checksum, 16) == chksum(pkt):
                    return pkt
                else:
                    raise Exception('Checksum Error')
                
            elif pkt_receiving == True:
                pkt += c


def spawn_gdb_server(self, target, port, do_forwarding=True, xml_file=None):
    if xml_file is None:
        # default for now: use ARM
        xml_file = f'{dirname(__file__)}/gdb/arm-target.xml'
      
    server = GDBRSPServer(self, target, port, xml_file, do_forwarding)
    server.start()
    self._gdb_servers.append(server)
    return server

def exit_server(avatar, watched_target):

    for s in avatar._gdb_servers:
        if s.target == watched_target:
            s.shutdown()
            avatar._gdb_servers.remove(s)

def load_plugin(avatar):
    avatar.spawn_gdb_server = MethodType(spawn_gdb_server, avatar)
    avatar.watchmen.add_watchman('TargetShutdown', when='before',
                                 callback=exit_server)
    avatar._gdb_servers = []
